/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "client.h"
#include "client_p.h"
#include "client_game.h"

// lib
#include <QFrustrationLib/protocol.h>

// Qt
#include <QtCore/QDebug>
#include <QtNetwork/QNetworkProxy>


namespace QFC {


ClientPrivate::ClientPrivate(Client *parent)
    : QTcpSocket(parent),
    q(parent)
{

    setProxy(QNetworkProxy::applicationProxy());

    connect(this, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(socketError(QAbstractSocket::SocketError)));
    connect(this, SIGNAL(readyRead()), this, SLOT(incomingData()));
    connect(this, SIGNAL(connected()), this, SLOT(tryLogin()));
    connect(this, SIGNAL(disconnected()), this, SLOT(disconnectedFromServer()));
    connect(this, SIGNAL(hostFound()), q, SIGNAL(hostFound()));
    connect(this, SIGNAL(proxyAuthenticationRequired(QNetworkProxy,QAuthenticator*)), q,
            SIGNAL(proxyAuthenticationRequired(QNetworkProxy,QAuthenticator*)));
    connect(this, SIGNAL(stateChanged(QAbstractSocket::SocketState)), q, SIGNAL(stateChanged(QAbstractSocket::SocketState)));

}


ClientPrivate::~ClientPrivate()
{



}


QFC::Game *ClientPrivate::getGameFromID(const int &id)
{

    for (int i = 0; i < games.size(); i++) {
        QFC::Game *game = games.at(i);
        if (game->gameId() == id) {
            return game;
        }
    }
    return 0;

}


void ClientPrivate::pong()
{

  //  qDebug() << "write pong";
    write(QF::Protocol::pong());

}


void ClientPrivate::socketError(const QAbstractSocket::SocketError &error)
{

    qDebug() << "QFC Client error:" << error;
    emit q->socketError(q, error, errorString());

}


static void corruptedData(const QF::Protocol::BlobType &type, const QStringList &args)
{

    qWarning() << "#######";
    qWarning() << "ERROR: Corrupted data from server:";
    qWarning() << QString("Blob Type: %1").arg(type);
    qWarning() << QString("Args: %2").arg(args.join(", "));
    qWarning() << "#######";

}



void ClientPrivate::incomingData()
{

    m_buffer.append(readAll());

    QByteArray blob;
    while (QF::Protocol::getBlobFromBuffer(&m_buffer, &blob)) {
        QStringList args;
        QF::Protocol::BlobType type;

        if (!QF::Protocol::parseBlob(blob, &type, &args)) {
            qWarning() << "ERROR: Parse Error, blob:" << blob;
            continue;
        }

       // qDebug() << "Got blob:" << blob;

        switch (type) {
        case QF::Protocol::PingType: pong(); break;
        case QF::Protocol::SetNickType: {
                if (QF::Protocol::parseServerSetsNick(&args, &nick)) {
                    emit q->connected(q);
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::StoneMovedType: {
                int from;
                int to;
                int gameID;

                if (QF::Protocol::parseStoneMoved(&args, &from, &to, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setStoneMoved(from, to);
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
        }
        case QF::Protocol::CreateGameType: {
                int gameID;
                QString name;
                QString key;
                QF::Layout layout;

                if (QF::Protocol::parseGameCreated(&args, &gameID, &name, &key, &layout)) {
                    QFC::Game *game = new QFC::Game(gameID, name, key, layout, true, q);
                    games.append(game);

                    emit q->gameCreated(q, game);
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::GameJoinedType: {
            QString playerName;
            int playerID;
            int gameID;
            QString gameName;
            QHash<int, QString> playerList;
            QF::Layout layout;

            if (QF::Protocol::parseGameJoined(&args, &playerName, &playerID, &gameID, &gameName, &playerList, &layout)) {
                QFC::Game *game = getGameFromID(gameID);
                if (game) {
                    game->setPlayerList(playerList);
                    game->setPlayerJoined(playerID, playerName);
                } else { // we joined
                    const QString key = keys.take(gameID);
                    game = new QFC::Game(gameID, gameName, key, layout, false, q);
                    game->setPlayerList(playerList);
                    games.append(game);
                    emit q->gameCreated(q, game);
                }
            } else {
                corruptedData(type, args);
            }
            break;
        }
        case QF::Protocol::GameLeftType: {
                int gameID;
                int playerID;
                QString nick;

                if (QF::Protocol::parseGameLeft(&args, &nick, &playerID, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setPlayerLeft(playerID, nick);
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::GameListType: {
                QHash<int, QString> list;
                if (QF::Protocol::parseGameList(&args, &list)) {
                    emit q->gotGameList(list);
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::GameOpenType: {
                int gameID;
                QString name;

                if (QF::Protocol::parseGameOpen(&args, &gameID, &name)) {
                    emit q->gameOpen(gameID, name);
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::GameClosedType: {
                int gameID;
                QString name;

                if (QF::Protocol::parseGameClosed(&args, &gameID, &name)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setGameClosed();
                    }
                    emit q->gameClosed(gameID, name);
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::StoneAddedType: {
                int playerID;
                int stoneID;
                int stonePos;
                int gameID;

                if (QF::Protocol::parseStoneAdded(&args, &playerID, &stoneID, &stonePos, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        QF::Stone *stone = new QF::Stone(game, playerID, stoneID, gameID);
                        stone->setPos(stonePos);
                        game->addStone(stone);
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::GamePlayerChangedType: {
                int playerID;
                int gameID;
                if (QF::Protocol::parseCurrentPlayerChanged(&args, &playerID, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setCurrentPlayer(playerID);
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::RollType: {
                int value;
                int gameID;

                if (QF::Protocol::parseRoll(&args, &value, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setRoll(value);
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::TurnType: {
                QHash<int, int> turn;
                int gameID;
                if (QF::Protocol::parseTurn(&args, &turn, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setTurn(turn);
                    }
                } else {
                    corruptedData(type, args);
                }
            break;
        }
        case QF::Protocol::GameChatType: {
                QString text;
                QString sender;
                int gameID;

                if (QF::Protocol::parseGameChat(&args, &text, &sender, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->appendChat(sender, text);
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::GameStartType: {
                int gameID;
                if (QF::Protocol::parseGameStarted(&args, &gameID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setGameStarted();
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::GameWinnerType: {
                int gameID;
                int playerID;

                if (QF::Protocol::parseGameWinner(&args, &gameID, &playerID)) {
                    QFC::Game *game = getGameFromID(gameID);
                    if (game) {
                        game->setWinner(playerID);
                    }
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        case QF::Protocol::ErrorType: {
                QF::Protocol::Error error;
                if (QF::Protocol::parseError(&args, &error)) {
                    emit q->error(q, error);
                } else {
                    corruptedData(type, args);
                }
                break;
            }
        default: qDebug() << "INVALID BLOB:" << blob;
        }
    }

}


void ClientPrivate::tryLogin()
{

    write(QF::Protocol::requestNick(nick));

}


void ClientPrivate::disconnectedFromServer()
{

    emit q->disconnected(q);

}


Client::Client(QObject *parent)
    : QObject(parent),
    d(new ClientPrivate(this))
{


}


Client::~Client()
{

    delete d;

}


QString Client::nick() const
{

    return d->nick;

}


QString Client::hostName() const
{

    return d->peerName();

}


quint16 Client::port() const
{

    return d->peerPort();

}


bool Client::isConnected() const
{

    switch (d->state()) {
    case QAbstractSocket::UnconnectedState:
    case QAbstractSocket::ClosingState: return false;
    default: return true;
    }

}


void Client::abort()
{

    d->abort();

}


void Client::connectToServer(const QString &hostName, const quint16 &port, const QString &nick)
{

    d->nick = nick;
    d->connectToHost(hostName, port, QTcpSocket::ReadWrite);

}


void Client::disconnectFromServer()
{

    d->disconnectFromHost();
    if (isConnected()) {
        d->waitForDisconnected();
    }

}


void Client::setNick(const QString &nick)
{

    d->write(QF::Protocol::requestNick(nick));
    d->flush();

}


void Client::waitForDisconnected(const int &msecs)
{

    d->waitForDisconnected(msecs);

}


void Client::requestRoll(const int &gameID)
{

    d->write(QF::Protocol::requestRoll(gameID));

}


void Client::requestTurn(const int &from, const int &to, const int &gameID)
{

    d->write(QF::Protocol::requestTurn(from, to, gameID));

}


void Client::requestGameList()
{

    d->write(QF::Protocol::requestGameList());

}


void Client::createGame(const QString &name, const QString &key, const QF::Layout &layout)
{

    d->write(QF::Protocol::sendCreateGame(name, key, layout));

}


void Client::joinGame(const int &gameID, const QString &key)
{

    d->keys.insert(gameID, key);
    d->write(QF::Protocol::requestJoinGame(gameID, key));

}


void Client::startGame(const int &gameID)
{

    d->write(QF::Protocol::requestStart(gameID));

}


void Client::gameChat(const QString &text, const int &gameID)
{

    d->write(QF::Protocol::sendGameChat(text, nick(), gameID));

}


void Client::leaveGame(const int &gameID)
{

    d->write(QF::Protocol::requestLeave(gameID));

}


void Client::gameLeft(QFC::Game *game)
{

    game->disconnect();
    d->games.removeAll(game);
    game->deleteLater();

}



} // namespace QFC


