/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "computerenemy.h"
#include "client_game.h"
#include "config.h"

// Qt
#include <QtCore/QDebug>
#include <QtCore/QTimer>
#include <QtCore/QDateTime>


#define DELAY m_delay
ComputerEnemy::ComputerEnemy(QObject *parent)
    : QObject(parent)
{

    m_gameID = -1;
    qsrand(QDateTime::currentDateTime().toTime_t());

    m_client = new QFC::Client(this);
    connect(m_client, SIGNAL(error(QFC::Client*,QF::Protocol::Error)), this, SLOT(error(QFC::Client*,QF::Protocol::Error)));
    connect(m_client, SIGNAL(gameCreated(QFC::Client*,QFC::Game*)), this, SLOT(gameCreated(QFC::Client*,QFC::Game*)));
    connect(m_client, SIGNAL(connected(QFC::Client*)), this, SLOT(connected(QFC::Client*)));
    connect(m_client, SIGNAL(disconnected(QFC::Client*)), this, SLOT(deleteLater()));
    connect(m_client, SIGNAL(destroyed()), this, SLOT(deleteLater()));

    connect(Config::self(), SIGNAL(configChanged()), this, SLOT(configChanged()));
    configChanged();

}


ComputerEnemy::~ComputerEnemy()
{

    delete m_client;

}


void ComputerEnemy::connectToServer(const QString &server, const quint16 &port)
{

    m_nick = tr("Computer")+'1';
    m_server = server;
    m_port = port;
    m_client->connectToServer(server, port, m_nick);

}


void ComputerEnemy::joinGame(const int &gameID, const QString &key)
{

    Q_ASSERT(m_gameID != -1);

    m_client->joinGame(gameID, key);
    m_gameID = gameID;

}


void ComputerEnemy::init(const QString &server, const quint16 &port, const int &gameID, const QString &key)
{

    m_port = port;
    m_key = key;
    m_gameID = gameID;
    m_server = server;

    connectToServer(server, port);

}


void ComputerEnemy::init(QFC::Game *game)
{

    m_nick = game->client()->nick();
    connect(game->client(), SIGNAL(error(QFC::Client*,QF::Protocol::Error)), this, SLOT(error(QFC::Client*,QF::Protocol::Error)));
    connect(game->client(), SIGNAL(disconnected(QFC::Client*)), this, SLOT(deleteLater()));
    connect(game->client(), SIGNAL(destroyed()), this, SLOT(deleteLater()));

    gameCreated(game->client(), game);

}


void ComputerEnemy::error(QFC::Client*, const QF::Protocol::Error &error)
{

    switch (error) {
    case QF::Protocol::NickAlreadyInUseErrorType:  {
            QRegExp rx("\\d+$");
            rx.indexIn(m_nick);
            const QString number = rx.cap();
          //  qDebug() << "increase number:" << number;
            m_nick.remove(rx);
            m_nick.append(QString::number(qMax(0, number.toInt())+1));
            m_client->setNick(m_nick);
            break;
        }
        default: qDebug() << "KI ERROR:" << m_nick << error; deleteLater(); break;
    }

}


void ComputerEnemy::gameCreated(QFC::Client *self, QFC::Game *game)
{

    Q_UNUSED(self);
    m_game = game;
    connect(game, SIGNAL(gotTurn(QHash<int,int>)), this, SLOT(gotTurn(QHash<int,int>)));
    connect(game, SIGNAL(currentPlayerChanged(int)), this, SLOT(playerChanged(int)));

    game->blockSignals(false);

}


void ComputerEnemy::gotTurn(const QHash<int, int> &turn)
{

    //qDebug() << "KI GOT TURN:" << turn;

    m_turn = turn;
    QTimer::singleShot(DELAY, this, SLOT(sendTurn()));

}


void ComputerEnemy::playerChanged(const int &playerID)
{

    if (m_game->playerId(m_nick) == playerID) {
      //  qDebug() << "KI PLAYER CHANGED";
        QTimer::singleShot(DELAY, this, SLOT(requestRoll()));
    }

}


void ComputerEnemy::connected(QFC::Client*)
{

    if (m_gameID != -1) {
        joinGame(m_gameID, m_key);
    }

}


bool ComputerEnemy::doIt(const int &percent) const
{

    return !(((qrand() % 100)+1) > percent);

}


bool ComputerEnemy::test(const Difficulty &d) const
{

    switch (d) {
    case VeryHard: return true;
    case Hard: return doIt(85);
    case Normal: return doIt(70);
    case Easy: return doIt(50);
    case VeryEasy: return doIt(5); break;
    }
    return false;

}


QF::Stone *ComputerEnemy::stoneOnIndex(const int &index) const
{

    foreach (QF::Stone *stone, m_game->stones()) {
        if (stone->pos() == index) {
            return stone;
        }
    }
    return 0;

}


void ComputerEnemy::sendTurn()
{

   // qDebug() << "KI SEND TURN";

    if (m_game->currentPlayer() != m_game->playerId(m_nick)) {
        qDebug() << "KI != CURRENT PLAYER";
        return;
    }

    if (m_turn.isEmpty()) {
        return;
    }

    QList<QF::Stone*> stones = m_game->stones();
    const int myId = m_game->playerList().key(m_nick);
    const QF::Layout layout = m_game->layout();
    QHash<int, PlayerData> data;

    // collect data
    for (int i = 0; i < layout.players(); i++) {
        const QList<int> home = layout.homes(i);
        const QList<int> targets = layout.targets(i);

        PlayerData d;
        foreach (QF::Stone *stone, stones) {
            if (stone->playerId() != i) {
                continue;
            }
            if (home.contains(stone->pos())) {
                d.inHome.append(stone);
            } else if (targets.contains(stone->pos())) {
                d.inTarget.append(stone);
            } else {
                d.onWay.append(stone);
            }
        }
        d.homeIndexList = home;
        d.targetIndexList = targets;
        d.start = layout.startField(i);
        d.last = layout.lastField(i);
        data.insert(i, d);
    }

    QHash<int, int> moves = m_turn;
    QHash<int, int> backup = moves;

    // kick somebody??
    if (test(m_difficulty)) {
        QHashIterator<int, int> it(m_turn);
        while (it.hasNext()) {
            it.next();

            for (int i = 0; i < stones.size(); i++) {
                QF::Stone *stone = stones.at(i);

                if (stone->pos() == it.value()) {
                    m_game->sendTurn(it.key(), it.value());
                    return;
                }
            }
        }
        // from way to target
        if (test(m_difficulty)) {
            QHashIterator<int, int> it(m_turn);
            while (it.hasNext()) {
                it.next();

                if (data.value(myId).targetIndexList.contains(it.value()) &&
                    !data.value(myId).targetIndexList.contains(it.key())) {
                    m_game->sendTurn(it.key(), it.value());
                    return;
                }
            }

            // from one target into another
            if (test(m_difficulty)) {
                for (int i = 0; i < stones.size(); i++) {
                    QF::Stone *stone = stones.at(i);

                    if (data.value(myId).targetIndexList.contains(stone->pos())
                        && m_turn.contains(stone->pos())) {
                        m_game->sendTurn(stone->pos(), m_turn.value(stone->pos()));
                        return;
                    }
                }
                // don't go on start fields from other players
                if (test(m_difficulty)) {
                    QMutableHashIterator<int, int> it(moves);
                    while (it.hasNext()) {
                        it.next();

                        for (int i = 0; i < layout.players(); i++) {
                            if (i == myId) {
                                continue;
                            }
                            if (it.value() == layout.startField(i) && !data.value(i).inHome.isEmpty()) {
                                it.remove();
                                break;
                            }
                        }
                    }
                    if (moves.isEmpty()) {
                        moves = backup;
                    }

                    backup = moves;

                    // stay behind the opponent
                    if (test(m_difficulty)) {
                        QMutableHashIterator<int, int> it(moves);
                        while (it.hasNext()) {
                            it.next();

                            if (data.value(myId).targetIndexList.contains(it.value()) || data.value(myId).homeIndexList.contains(it.key())) {
                                continue;
                            }
                            int count = layout.getWay(myId, it.key(), it.value()).size();
                            for (int i = 1; i < count; i++) {
                                int index = layout.tryMove(myId, it.key(), i);
                                Q_ASSERT(index != -1);

                                QF::Stone *stone = stoneOnIndex(index);
                                if (stone && stone->playerId() != myId) {
                                    it.remove();
                                    break;
                                }
                            }
                        }
                        if (moves.isEmpty()) {
                            moves = backup;
                        }

                        backup = moves;


                        // -------------------------
                        // ok, choose the best stone
                        // -------------------------

                        // stone at the start of an opponent
                        if (test(m_difficulty)) {
                            QHash<int, int> result;
                            QHashIterator<int, int> it(moves);
                            while (it.hasNext()) {
                                it.next();

                                for (int i = 0; i < layout.players(); i++) {
                                    if (i == myId) {
                                        continue;
                                    }
                                    if (data.value(i).start == it.key()) {
                                        result.insert(it.key(), it.value());
                                        break;
                                    }
                                }
                            }
                            if (!result.isEmpty()) {
                                moves = result;
                            }

                            // closest to the target
                            if (test(m_difficulty)) {
                                const PlayerData d = data.value(myId);
                                int fieldCount = 100000;
                                QPair<int, int> best = qMakePair(-1, -1);
                                QHashIterator<int, int> it(moves);
                                while (it.hasNext()) {
                                    it.next();


                                    if (d.homeIndexList.contains(it.key()) ||
                                        d.targetIndexList.contains(it.key()) ||
                                        d.targetIndexList.contains(it.value()))
                                    {
                                        continue;
                                    }

                                    const QList<int> way = layout.getWay(myId, it.key(), d.last);
                                    Q_ASSERT(!way.isEmpty());
                                    if (way.size() < fieldCount) {
                                        fieldCount = way.size();
                                        best.first = it.key();
                                        best.second = it.value();
                                    }
                                }

                                if (best.first != -1) {
                                    moves.clear();
                                    moves.insert(best.first, best.second);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    if (moves.isEmpty()) {
        moves = backup;
    }

    Q_ASSERT(!moves.isEmpty());

    // random move
    QHash<int, int>::iterator i = moves.begin();
    m_game->sendTurn(i.key(), i.value());

}


void ComputerEnemy::requestRoll()
{

    if (m_game->currentPlayer() == m_game->playerId(m_nick)) {
      //  qDebug() << "KI REQUEST ROLL";
        m_game->roll();
    } else {
        qDebug() << "WARNING REQUEST ROLL"<< m_game->currentPlayer() << m_game->playerId(m_nick) << m_nick;
    }

}


void ComputerEnemy::configChanged()
{

    m_delay = Config::self()->kiDelay();
    m_difficulty = (Difficulty) Config::self()->value("KI Difficulty", 3);

}


