/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// own
#include "config.h"
#include "theme.h"
#include "themeconfigpage.h"
#include "chatconfigpage.h"
#include "keysconfigpage.h"
#include "generalconfigpage.h"
#include "soundconfigpage.h"
#include "proxyconfigpage.h"
#include "soundcore.h"

// lib
#include <QFrustrationSharedGUI/global.h>

// Qt
#include <QtCore/QSettings>
#include <QtCore/QStringList>
#include <QtCore/QDebug>
#include <QtNetwork/QNetworkProxy>


Config::~Config()
{

    save();
    delete m_settings;

}


class ConfigSingleton
{
public:
    Config self;
};


FRUSTRATION_SINGLETON(ConfigSingleton, configSelf)


Config *Config::self()
{

    return &configSelf->self;

}


SoundCore *Config::sound() const
{

    return m_sound;

}


void Config::setSound(SoundCore *sound)
{

    m_sound = sound;
    updateSounds();

}


bool Config::isWritable() const
{

    return m_settings->isWritable();

}


QString Config::configPath() const
{

    return m_settings->fileName();

}


int Config::kiDelay() const
{

    return m_data.value("kiDelay").toInt();

}


int Config::animationSpeed() const
{

    return m_data.value("animationSpeed").toInt();

}


bool Config::openGL() const
{

    return m_data.value("OpenGL").toBool();

}


Qt::Key Config::key(const Config::KeyType &type) const
{

    return (Qt::Key) m_data.value(QString("Key %1").arg(type)).toInt();

}


bool Config::isChatMessageEnabled(const Config::ChatMessageType &type) const
{

    return m_data.value(QString("ChatMessage%1").arg((int) type)).toBool();

}


QVariant Config::value(const QString &key) const
{

    return m_data.value(key);

}


void Config::setKey(const Config::KeyType &type, const Qt::Key &key)
{

    m_data.insert(QString("Key %1").arg(type), (int) key);

}


void Config::setKIDelay(const int &delay)
{

    m_data.insert("kiDelay", delay);

}


void Config::setAnimationSpeed(const int &speed)
{

    m_data.insert("animationSpeed", speed);

}


void Config::setOpenGL(const bool &enable)
{

    m_data.insert("OpenGL", enable);

}


void Config::setChatMessageEnabled(const Config::ChatMessageType &type, const bool &enabled)
{

    m_data.insert(QString("ChatMessage%1").arg((int) type), enabled);

}


void Config::setValue(const QString &key, const QVariant &value)
{

    m_data.insert(key, value);

}


void Config::saveProxy()
{

    const QNetworkProxy proxy = QNetworkProxy::applicationProxy();
    m_data.insert("Proxy Host", proxy.hostName());
    m_data.insert("Proxy Port", (int) proxy.port());
    m_data.insert("Proxy User", proxy.user());
    m_data.insert("Proxy Password", proxy.password()); // FIXME

}


void Config::loadProxy()
{

    QNetworkProxy proxy = QNetworkProxy::applicationProxy();
    proxy.setHostName(m_data.value("Proxy Host").toString());
    proxy.setPort(m_data.value("Proxy Port").toInt());
    proxy.setUser(m_data.value("Proxy User").toString());
    proxy.setPassword(m_data.value("Proxy Password").toString());
    QNetworkProxy::setApplicationProxy(proxy);

}


void Config::sendChanged()
{

    emit configChanged();

}


Config::Config(QObject *parent)
    : QObject(parent)
{

    m_settings = new QSettings(this);

    m_sound = 0;

    load();
    connect(this, SIGNAL(configChanged()), this, SLOT(updateSounds()));

}


void Config::load()
{

    m_data.clear();

    const QStringList keys = m_settings->allKeys();
    for (int i = 0; i < keys.size(); i++) {
        const QString key = keys.at(i);

        m_data.insert(key, m_settings->value(key, QVariant()));
    }

    if (!m_data.contains("FirstStart")) {
        m_data.insert("FirstStart", false);

        GeneralConfigPage::saveDefaults(this);
        KeysConfigPage::saveDefaults(this);
        ChatConfigPage::saveDefaults(this);
        SoundConfigPage::saveDefaults(this);
        ThemeConfigPage::saveDefaults(this);
    }

    Theme::currentTheme()->setCurrentTheme(m_settings->value("Theme", "Frusti").toString());

}


void Config::save()
{

    QHashIterator<QString, QVariant> it(m_data);
    while (it.hasNext()) {
        it.next();

        m_settings->setValue(it.key(), it.value());
    }
    m_settings->setValue("Theme", Theme::currentTheme()->currentThemeName());
    m_settings->sync();

}


void Config::updateSounds()
{

    if (!m_sound) {
        return;
    }
    m_sound->addSound("Dice", value("Sound-Dice").toString());
    m_sound->addSound("Winner", value("Sound-Winner").toString());
    m_sound->addSound("Loser", value("Sound-Loser").toString());
    m_sound->addSound("PlayerJoined", value("Sound-PlayerJoined").toString());
    m_sound->addSound("PlayerLeft", value("Sound-PlayerLeft").toString());

}

