/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// own
#include "gametab.h"
#include "ui_gametab.h"
#include "client.h"
#include "computerenemy.h"
#include "keydialog.h"
#include "playerlistdock.h"
#include "chatdock.h"
#include "config.h"
#include "soundcore.h"

// lib
#include <QFrustrationSharedGUI/global.h>
#include <QFrustrationSharedGUI/errordialog.h>

// Qt
#include <QtCore/QDebug>
#include <QtGui/QToolBar>
#include <QtGui/QMessageBox>
#include <QtGui/QAction>
#include <QtCore/QTimer>


GameTab::GameTab(QWidget *parent)
    : QMainWindow(parent),
    ui(new Ui::GameTab)
{

    m_ki = 0;

    setCentralWidget(new QWidget(this));
    ui->setupUi(centralWidget());
    m_currentPlayer = 0;

    connect(ui->view, SIGNAL(diceClicked()), this, SLOT(diceClicked()));
    connect(ui->view, SIGNAL(sendTurn(int,int)), this, SLOT(sendTurn(int, int)));

    QAction *startGameAction = getAction("start_action");
    startGameAction->setText(tr("Start"));
    startGameAction->setIcon(FRUST_ICON("system-run"));
    connect(startGameAction, SIGNAL(triggered()), this, SLOT(startClicked()));

    QAction *zoomInAction = getAction("zoom_in_action");
    zoomInAction->setShortcut(QKeySequence::ZoomIn);
    zoomInAction->setText(tr("Zoom in"));
    zoomInAction->setIcon(FRUST_ICON("zoom-in"));
    connect(zoomInAction, SIGNAL(triggered()), this, SLOT(zoomInClicked()));

    QAction *zoomOutAction = getAction("zoom_out_action");
    zoomOutAction->setShortcut(QKeySequence::ZoomOut);
    zoomOutAction->setText(tr("Zoom out"));
    zoomOutAction->setIcon(FRUST_ICON("zoom-out"));
    connect(zoomOutAction, SIGNAL(triggered()), this, SLOT(zoomOutClicked()));

    QAction *addComputerPlayer = getAction("add_ki_action");
    addComputerPlayer->setText(tr("Add Computer Player"));
    addComputerPlayer->setIcon(FRUST_ICON("computer"));
    connect(addComputerPlayer, SIGNAL(triggered()), this, SLOT(addComputerPlayerClicked()));

    QAction *fullViewAction = getAction("full_view_action");
    fullViewAction->setCheckable(true);
    fullViewAction->setChecked(false);
    connect(fullViewAction, SIGNAL(triggered(bool)), this, SLOT(fullViewTriggered(bool)));
    fullViewAction->setText(tr("Maximize Tab"));
    fullViewAction->setIcon(FRUST_ICON("view-fullscreen"));


    QToolBar *bar = new QToolBar(tr("Game Toolbar"), this);
    bar->setObjectName("GameTabToolBar");

    bar->addAction(startGameAction);
    bar->addSeparator();
    bar->addAction(zoomInAction);
    bar->addAction(zoomOutAction);
    bar->addSeparator();
    bar->addAction(addComputerPlayer);
    bar->addSeparator();
    bar->addAction(fullViewAction);

    addToolBar(Qt::TopToolBarArea, bar);


    m_playerList = new PlayerListDock(this);
    m_chatDock = new ChatDock(this);

    QTimer::singleShot(0, this, SLOT(init()));

}


GameTab::~GameTab()
{

    if (getAction("full_view_action")->isChecked()) {
        emit fullViewModeTriggered(this, false);
    }

    Config::self()->setValue("GameTabState", saveState());

    delete ui;
    if (m_game) {
        m_game.data()->leave();
        m_game.data()->client()->gameLeft(m_game.data());
    }

}


void GameTab::init()
{

    addDockWidget(Qt::LeftDockWidgetArea, m_playerList);
    addDockWidget(Qt::LeftDockWidgetArea, m_chatDock);
    restoreState(Config::self()->value("GameTabState").toByteArray());
    connect(m_chatDock, SIGNAL(sendChat(QString)), this, SLOT(sendChat(QString)));

}


QString GameTab::title() const
{

    if (m_game) {
        return m_game.data()->name();
    } else {
        return QString();
    }

}


void GameTab::setGame(QFC::Game *game)
{

    qDebug() << "got new game;" << game->gameId();

    m_game = game;

    QHash<int, QString> playerList = game->playerList();
    QHashIterator<int, QString> it(playerList);
    while (it.hasNext()) {
        it.next();
        m_playerList->addPlayer(it.value(), it.key());
    }
    m_playerList->setSelf(playerList.key(game->client()->nick()));
    ui->view->setSelf(playerList.key(game->client()->nick()));

    connect(game, SIGNAL(playerJoined(QString,int)), this, SLOT(playerJoined(QString,int)));
    connect(game, SIGNAL(playerLeft(QString,int)), this, SLOT(playerLeft(QString,int)));
    connect(game, SIGNAL(stoneAdded(QF::Stone*)), this, SLOT(stoneAdded(QF::Stone*)));
    connect(game, SIGNAL(currentPlayerChanged(int)), this, SLOT(currentPlayerChanged(int)));
    connect(game, SIGNAL(diceRolled(int)), this, SLOT(diceRolled(int)));
    connect(game, SIGNAL(gotTurn(QHash<int,int>)), this, SLOT(gotTurn(QHash<int,int>)));
    connect(game, SIGNAL(newChatText(QString,QString)), this, SLOT(newChatText(QString,QString)));
    connect(game, SIGNAL(gameStarted()), this, SLOT(gameStarted()));
    connect(game, SIGNAL(winner(int)), this, SLOT(winner(int)));
    connect(game, SIGNAL(gameClosed()), this, SLOT(gameClosed()));

    ui->view->setLayout(game->layout());
    ui->view->setPlayerList(playerList);
    game->blockSignals(false);

    // for tests only
    //m_ki = new ComputerEnemy(this);
    //m_ki->init(game);

}


void GameTab::startQuickGame(const int &kiCount)
{

    for (int i = 0; i < kiCount; i++) {
        addComputerPlayerClicked();
    }

}


QAction *GameTab::getAction(const QString &name)
{

    for (int i = 0; i < m_actions.size(); i++) {
        QAction *action = m_actions.at(i);
        if (action->objectName() == name) {
            return action;
        }
    }
    QAction *action = new QAction(this);
    action->setObjectName(name);
    m_actions.append(action);
    return action;

}


void GameTab::playerJoined(const QString &nick, const int &id)
{

    m_playerList->addPlayer(nick, id);
    m_chatDock->addPlayer(nick, id);
    if (m_game) {
        ui->view->setPlayerList(m_game.data()->playerList());
    }
    Config::self()->sound()->play("PlayerJoined");

}


void GameTab::playerLeft(const QString &nick, const int &id)
{

    m_playerList->removePlayer(nick, id);
    m_chatDock->removePlayer(nick, id);
    if (m_game) {
        ui->view->setPlayerList(m_game.data()->playerList());
    }
    Config::self()->sound()->play("PlayerLeft");

}


void GameTab::startClicked()
{

    if (m_game) {
        m_game.data()->requestStart();
    }

}


void GameTab::stoneAdded(QF::Stone *stone)
{

    ui->view->addStone(stone);

}


void GameTab::currentPlayerChanged(const int &newCurrentPlayerID)
{

    ui->view->setCurrentPlayer(newCurrentPlayerID);
    m_currentPlayer = newCurrentPlayerID;
    m_playerList->setCurrentPlayer(newCurrentPlayerID);

}


void GameTab::diceRolled(const int &value)
{

    Q_ASSERT(m_game);

    if (m_game) {
        ui->view->setRoll(value);
        m_chatDock->setRoll(m_game.data()->playerList().value(m_currentPlayer), value);
    }

}


void GameTab::gotTurn(const QHash<int, int> &turn)
{

    if (!m_ki) {
        ui->view->setTurn(turn);
    }

}


void GameTab::diceClicked()
{

    if (m_game) {
        m_game.data()->roll();
    }

}


void GameTab::sendTurn(const int &from, const int &to)
{

    qDebug() << "send turn:" << from << to;
    if (m_game) {
        m_game.data()->sendTurn(from, to);
    }

}


void GameTab::newChatText(const QString &sender, const QString &text)
{

    m_chatDock->appendChat(sender, text);

}


void GameTab::sendChat(const QString &text)
{

    if (m_game) {
        m_game.data()->sendChat(text);
    }

}


void GameTab::gameStarted()
{

    getAction("start_action")->setEnabled(false);

}


void GameTab::winner(const int &playerID)
{

    if (m_game) {
        if (m_game.data()->playerList().key(m_game.data()->client()->nick()) == playerID) {
            Config::self()->sound()->play("Winner");
        } else {
            Config::self()->sound()->play("Loser");
        }

        m_chatDock->appendChat(tr("Server"), tr("\"%1\" won!").arg(m_game.data()->playerList().value(playerID)));
    }
    ui->view->setDisabled(true);

}


void GameTab::gameClosed()
{

    if (m_game && m_game.data()->hasWinner()) {
        return;
    }

    ErrorDialog::showError(parentWidget(), // this tab will be deleted
                           tr("Game Stopped"),
                           tr("There are too few players left, so the game must be stopped."));

    emit leaveGame(this);

}


void GameTab::zoomInClicked()
{

    const int zoom = ui->view->currentZoom()+10;
    ui->view->setZoom(zoom);

}


void GameTab::zoomOutClicked()
{

    const int zoom = ui->view->currentZoom()-10;
    ui->view->setZoom(zoom);

}


void GameTab::addComputerPlayerClicked()
{

    if (!m_game) {
        return;
    }

    const QString key = m_game.data()->key();
    const int id = m_game.data()->gameId();
    const QString host = m_game.data()->client()->hostName();
    const qint16 port = m_game.data()->client()->port();

    ComputerEnemy *e = new ComputerEnemy(this);
    e->init(host, port, id, key);

}


void GameTab::fullViewTriggered(const bool &checked)
{

    emit fullViewModeTriggered(this, checked);

}



