/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// onw
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "client.h"
#include "serverdialog.h"
#include "configdialog.h"
#include "startserverdialog.h"
#include "localservertab.h"
#include "config.h"
#include "servertab.h"
#include "gametab.h"
#include "soundcore.h"
#include "connectingdialog.h"

// lib
#include <QFrustrationSharedGUI/global.h>
#include <QFrustrationSharedGUI/aboutdialog.h>
#include <QFrustrationSharedGUI/errordialog.h>
#include <QFrustrationSharedGUI/questiondialog.h>
#include <QFrustrationLib/server.h>

// Qt
#include <QtCore/QDebug>
#include <QtCore/QDateTime>
#include <QtCore/QTimer>
#include <QtGui/QAction>
#include <QtGui/QStackedLayout>
#include <QtGui/QApplication>


MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{

    if (!Config::self()->isWritable()) {
        ErrorDialog::showError(this,
                               tr("Warning"),
                               tr("%1 does not have the necessary rights "
                                  "to store its configuration."
                                  "The configuration is stored in %2."
                                  "You must fix the problem,"
                                  "if you want to save the configuration.")
                               .arg(qApp->applicationName()).arg(Config::self()->configPath()));
    }

    Config::self()->setSound(new SoundCore(this));
    Config::self()->loadProxy();


    m_serverTab = 0;
    m_localServerTab = 0;

    ui->setupUi(this);
    ui->mainLayout->removeWidget(ui->serverTabWidget);

    m_layout = new QStackedLayout;
    m_layout->setContentsMargins(0, 0, 0, 0);
    m_layout->addWidget(ui->serverTabWidget);
    m_layout->setCurrentWidget(ui->serverTabWidget);
    ui->mainLayout->addLayout(m_layout);

    connect(ui->serverTabWidget, SIGNAL(tabCloseRequested(int)), this, SLOT(serverTabCloseRequested(int)));

    QTimer::singleShot(0, this, SLOT(setupActions()));

    setMinimumSize(200, 200);
    resize(600, 400);

    restoreState(Config::self()->value("MainWindowState", QByteArray()));
    restoreGeometry(Config::self()->value("MainWindowGeometry", QByteArray()));

    m_client = new QFC::Client(this);
    connect(m_client, SIGNAL(connected(QFC::Client*)), this, SLOT(clientConnected(QFC::Client*)));
    connect(m_client, SIGNAL(error(QFC::Client*,QF::Protocol::Error)), this, SLOT(gotErrorReply(QFC::Client*,QF::Protocol::Error)));
    connect(m_client, SIGNAL(gameCreated(QFC::Client*,QFC::Game*)), this, SLOT(newGame(QFC::Client*,QFC::Game*)));
    connect(m_client, SIGNAL(socketError(QFC::Client*,QAbstractSocket::SocketError,QString)), this,
            SLOT(socketError(QFC::Client*,QAbstractSocket::SocketError,QString)));

}


MainWindow::~MainWindow()
{

    Config::self()->setValue("MainWindowState", saveState());
    Config::self()->setValue("MainWindowGeometry", saveGeometry());

    if (m_serverTab) {
        clientDisconnected(m_serverTab);
    }

    delete ui;
    delete m_client;

    Config::self()->sound()->deleteLater();

}


void MainWindow::setupActions()
{

    QAction *startServerPrivateAction = new QAction(this);
    startServerPrivateAction->setShortcut(QKeySequence(Qt::CTRL+Qt::SHIFT+Qt::Key_S));
    startServerPrivateAction->setText(tr("Start Server (Private) ..."));
    startServerPrivateAction->setIcon(FRUST_ICON("network-server"));
    connect(startServerPrivateAction, SIGNAL(triggered()), this, SLOT(startServerPrivateClicked()));

    QAction *startServerPublicAction = new QAction(this);
    startServerPublicAction->setShortcut(QKeySequence(Qt::CTRL+Qt::SHIFT+Qt::Key_P));
    startServerPublicAction->setText(tr("Start Server (Public) ..."));
    startServerPublicAction->setIcon(FRUST_ICON("network-server"));
    connect(startServerPublicAction, SIGNAL(triggered()), this, SLOT(startServerPublicClicked()));

    QAction *connectToServerAction = new QAction(this);
    connectToServerAction->setShortcut(QKeySequence(Qt::CTRL+Qt::SHIFT+Qt::Key_C));
    connectToServerAction->setText(tr("Connect to Server ..."));
    connectToServerAction->setIcon(FRUST_ICON("network-connect"));
    connect(connectToServerAction, SIGNAL(triggered()), this, SLOT(connectToServerClicked()));

    QAction *quitAction = new QAction(this);
    quitAction->setShortcut(QKeySequence::Quit);
    quitAction->setText(tr("Quit"));
    quitAction->setIcon(FRUST_ICON("application-exit"));
    connect(quitAction, SIGNAL(triggered()), qApp, SLOT(quit()));

    QAction *fullScreenAction = new QAction(this);
    fullScreenAction->setShortcut(QKeySequence(Qt::CTRL+Qt::SHIFT+Qt::Key_F));
    fullScreenAction->setCheckable(true);
    fullScreenAction->setChecked(isFullScreen());
    fullScreenAction->setText(tr("Fullscreen"));
    fullScreenAction->setIcon(FRUST_ICON("view-fullscreen"));
    connect(fullScreenAction, SIGNAL(triggered(bool)), this, SLOT(fullScreenModeTriggered(bool)));

    QAction *configureAction = new QAction(this);
    configureAction->setText(tr("Configure %1 ...").arg(qApp->applicationName()));
    configureAction->setIcon(FRUST_ICON("configure"));
    connect(configureAction, SIGNAL(triggered()), this, SLOT(configure()));

    QAction *aboutAction = new QAction(this);
    aboutAction->setText(tr("About %1 ...").arg(qApp->applicationName()));
    aboutAction->setIcon(windowIcon());
    connect(aboutAction, SIGNAL(triggered()), this, SLOT(showAbout()));

    QAction *aboutQtAction = new QAction(this);
    aboutQtAction->setText(tr("About %1 ...").arg("Qt"));
    aboutQtAction->setIcon(FRUST_ICON("dialog-information"));
    connect(aboutQtAction, SIGNAL(triggered()), qApp, SLOT(aboutQt()));


    QMenu *gameMenu = new QMenu(this);
    gameMenu->setTitle(tr("Game"));

    gameMenu->addAction(connectToServerAction);
    gameMenu->addSeparator();
    gameMenu->addAction(startServerPrivateAction);
    gameMenu->addAction(startServerPublicAction);
    gameMenu->addSeparator();
    gameMenu->addAction(quitAction);


    QMenu *viewMenu = new QMenu(this);
    viewMenu->setTitle(tr("View"));

    viewMenu->addAction(fullScreenAction);


    QMenu *settingsMenu = new QMenu(this);
    settingsMenu->setTitle(tr("Settings"));

    settingsMenu->addAction(configureAction);


    QMenu *helpMenu = new QMenu(this);
    helpMenu->setTitle(tr("Help"));

    helpMenu->addAction(aboutQtAction);
    helpMenu->addAction(aboutAction);


    menuBar()->addMenu(gameMenu);
    menuBar()->addMenu(viewMenu);
    menuBar()->addMenu(settingsMenu);
    menuBar()->addMenu(helpMenu);

}


void MainWindow::serverDialogFinished(const QString &server, const quint16 &port, const QString &nick)
{

    ConnectingDialog::start(this, m_client);
    m_client->connectToServer(server, port, nick);

}


void MainWindow::clientConnected(QFC::Client *self)
{

    m_serverTab = new ServerTab(this);
    connect(m_serverTab, SIGNAL(disconnected(ServerTab*)), this, SLOT(clientDisconnected(ServerTab*)));
    m_serverTab->setServer(m_client);
    ui->serverTabWidget->addTab(m_serverTab, FRUST_ICON("network-server"), tr("%1: Game Room").arg(self->hostName()));
    ui->serverTabWidget->setCurrentWidget(m_serverTab);

}


void MainWindow::newGame(QFC::Client *server, QFC::Game *game)
{

    Q_UNUSED(server);
    Q_ASSERT(m_serverTab);

    GameTab *tab = new GameTab(this);
    connect(tab, SIGNAL(leaveGame(QWidget*)), this, SLOT(gameLeaved(QWidget*)));
    connect(tab, SIGNAL(fullViewModeTriggered(GameTab*,bool)), this, SLOT(fullViewModeTriggered(GameTab*,bool)));
    tab->setGame(game);
    ui->serverTabWidget->addTab(tab, FRUST_ICON("applications-games"), game->name());
    ui->serverTabWidget->setCurrentWidget(tab);

    m_serverTab->addGameTab(tab);

}


void MainWindow::gameLeaved(QWidget *tab)
{

    Q_ASSERT(m_serverTab);
    m_serverTab->removeGameTab(tab);
    ui->serverTabWidget->removeTab(ui->serverTabWidget->indexOf(tab));
    delete tab;

}


void MainWindow::socketError(QFC::Client *self, const QAbstractSocket::SocketError &error, const QString &errorString)
{

    Q_UNUSED(self);

    QString text = errorString;
    switch (error) {
    case QAbstractSocket::ConnectionRefusedError: text = tr("The connection was refused by the peer."); break;
    case QAbstractSocket::RemoteHostClosedError: text = tr("The remote host closed the connection."); break;
    case QAbstractSocket::HostNotFoundError: text = tr("The host address was not found."); break;
    case QAbstractSocket::SocketAccessError:
        text = tr("The socket operation failed because \"%1\" lacked the required privileges.").arg(qApp->applicationName());
        break;
    case QAbstractSocket::SocketTimeoutError: text = tr("Timeout."); break;
    case QAbstractSocket::NetworkError: text = tr("An error occurred with the network."); break;
    case QAbstractSocket::ProxyAuthenticationRequiredError: text = tr("The proxy requires authentication."); break;
    case QAbstractSocket::ProxyConnectionRefusedError:
        text = tr("Could not contact the proxy server because the connection to that server was denied.");
        break;
    case QAbstractSocket::ProxyConnectionTimeoutError: text = tr("The connection to the proxy server timed out."); break;
    case QAbstractSocket::ProxyNotFoundError: text = tr("The proxy address was not found."); break;
    default: break;
    }

    if (!text.isEmpty()) {
        ErrorDialog::showError(this, tr("Socket Error"), text);
    } else {
        qDebug() << "Unkown error:" << error;
    }

}


void MainWindow::gotErrorReply(QFC::Client *self, const QF::Protocol::Error &error)
{

    QString errorText;
    switch (error) {
    case QF::Protocol::TooManyPlayersErrorType: errorText = tr("There are already too many players in the game!"); break;
    case QF::Protocol::NotGameOwnerErrorType: errorText = tr("You are not the lead player!"); break;
    case QF::Protocol::NotEnoughPlayersErrorType: errorText = tr("There are not enough players present!"); break;
    case QF::Protocol::NotYourTurnErrorType: errorText = tr("It's not your turn!"); break;
    case QF::Protocol::InvalidTurnErrorType: qFatal("Invalid turn!"); break; // protocol error
    case QF::Protocol::GameAlreadyStartedErrorType: errorText = tr("The game has already started!"); break;
    case QF::Protocol::GameNotFoundErrorType: qFatal("Game not found!"); break; // protocol error
    case QF::Protocol::WrongKeyErrorType: errorText = tr("The password is incorrect!"); break;
    case QF::Protocol::GameNotStartedErrorType: errorText = tr("The game has not started!"); break;
    case QF::Protocol::NoNickSetErrorType: qFatal("No nick set!"); break; // protocol error
    case QF::Protocol::NickAlreadyInUseErrorType: {
            self->disconnectFromServer();
            errorText = tr("The nickname is already in use, please try again.");
            break;
        }
    case QF::Protocol::AlreadyInGameErrorType: errorText = tr("You are already in the game."); break;
    case QF::Protocol::PlayerHasAlreadyRolledErrorType: errorText = tr("You have already rolled."); break;
    case QF::Protocol::PlayerHasNotRolledErrorType: errorText = tr("You must first roll the dice."); break;
    case QF::Protocol::InvalidDataErrorType: qFatal("Protocol Error!"); break;
    }

    ErrorDialog::showError(this, tr("Error"), errorText);

}


void MainWindow::connectToServerClicked()
{

    if (m_client->isConnected()) {
        QuestionDialog *dialog = QuestionDialog::showQuestion(this,
                                                              tr("Disconnect"),
                                                              tr("You are already connected to the server "
                                                                 "\"%1\", you must close the connection first.\n"
                                                                 "Do you want to close the connection now?").arg(m_client->hostName()));
        connect(dialog, SIGNAL(yesClicked()), this, SLOT(showConnectToServerDialog()));
    } else {
        showConnectToServerDialog();
    }

}


void MainWindow::showConnectToServerDialog()
{

    if (m_client->isConnected()) {
        if (m_serverTab) {
            ui->serverTabWidget->removeTab(ui->serverTabWidget->indexOf(m_serverTab));
            delete m_serverTab;
            m_serverTab = 0;
        } else {
            m_client->abort();
        }
    }

    int port = 8080;
    if (m_localServerTab && m_localServerTab->server()) {
        port = m_localServerTab->server()->port();
    }

    ServerDialog *dialog = new ServerDialog("localhost", port, this);
    connect(dialog, SIGNAL(finished(QString, quint16, QString)), this, SLOT(serverDialogFinished(QString, quint16, QString)));

    dialog->show();
    dialog->raise();

}


void MainWindow::startServerPrivateClicked()
{

    if (m_localServerTab && m_localServerTab->server()) {
        QuestionDialog *dialog = QuestionDialog::showQuestion(this,
                                                              tr("Warning"),
                                                              tr("A server is already running, you must stop it first."
                                                                 "Do you want to stop the server now?"));
        connect(dialog, SIGNAL(yesClicked()), this, SLOT(startServerPrivate()));
    } else {
        startServerPrivate();
    }

}


void MainWindow::startServerPrivate()
{

    if (!m_localServerTab) {
        m_localServerTab = new LocalServerTab(this);
        ui->serverTabWidget->addTab(m_localServerTab, FRUST_ICON("network-server"), tr("Local Server"));
    }

    StartServerDialog *dialog = new StartServerDialog(false, m_localServerTab, this);
    dialog->show();
    dialog->raise();

}


void MainWindow::startServerPublicClicked()
{

    if (m_localServerTab && m_localServerTab->server()) {
        QuestionDialog *dialog = QuestionDialog::showQuestion(this,
                                                              tr("Warning"),
                                                              tr("A server is already running, you must stop it first."
                                                                 "Do you want to stop the server now?"));
        connect(dialog, SIGNAL(yesClicked()), this, SLOT(startServerPublic()));
    } else {
        startServerPublic();
    }

}


void MainWindow::startServerPublic()
{

    if (!m_localServerTab) {
        m_localServerTab = new LocalServerTab(this);
        ui->serverTabWidget->addTab(m_localServerTab, FRUST_ICON("network-server"), tr("Local Server"));
    }

    StartServerDialog *dialog = new StartServerDialog(true, m_localServerTab, this);
    dialog->show();
    dialog->raise();

}


void MainWindow::serverTabCloseRequested(const int &index)
{

    QWidget *widget = ui->serverTabWidget->widget(index);
    if (m_localServerTab && widget == m_localServerTab) {
        m_localServerTab = 0;
    } else if (m_serverTab && m_serverTab == widget) {
        foreach (QWidget *tab, m_serverTab->gameTabs()) {
            m_serverTab->removeGameTab(tab);
            ui->serverTabWidget->removeTab(ui->serverTabWidget->indexOf(tab));
            delete tab;
        }
        m_serverTab = 0;
    } else {
        Q_ASSERT(m_serverTab);
        m_serverTab->removeGameTab(widget);
    }

    ui->serverTabWidget->removeTab(index);
    delete widget;

}


void MainWindow::clientDisconnected(ServerTab *self)
{

    if (m_serverTab) {
        foreach (QWidget *tab, self->gameTabs()) {
            self->removeGameTab(tab);
            ui->serverTabWidget->removeTab(ui->serverTabWidget->indexOf(tab));
            delete tab;
        }
        m_serverTab = 0;
        ui->serverTabWidget->removeTab(ui->serverTabWidget->indexOf(self));
        delete self;
    }

}


void MainWindow::configure()
{

    ConfigDialog *dialog = new ConfigDialog(this);
    dialog->show();
    dialog->raise();

}


void MainWindow::showAbout()
{

    QHash<QString, QString> devs;
    devs.insert("Kai Dombrowe", "just89@gmx.de");
    AboutDialog::showAbout(this,
                           devs,
                           QUrl("https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=CGJG8DG4H72W2"));

}


void MainWindow::fullScreenModeTriggered(const bool &enabled)
{

    if (!enabled) {
        showNormal();
    } else {
        showFullScreen();
    }

}


void MainWindow::fullViewModeTriggered(GameTab *tab, const bool &enabled)
{

    if (enabled) {
        menuBar()->hide();
        centralWidget()->layout()->setContentsMargins(0, 0, 0, 0);

        ui->serverTabWidget->removeTab(ui->serverTabWidget->indexOf(tab));
        m_layout->addWidget(tab);
        m_layout->setCurrentWidget(tab);
    } else {
        menuBar()->show();
        centralWidget()->layout()->setContentsMargins(4, 4, 4, 4);

        m_layout->removeWidget(tab);
        m_layout->setCurrentWidget(ui->serverTabWidget);
        ui->serverTabWidget->addTab(tab, FRUST_ICON("applications-games"), tab->title());
        ui->serverTabWidget->setCurrentWidget(tab);
    }

}

