/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "servertab.h"
#include "ui_servertab.h"
#include "keydialog.h"
#include "creategamedialog.h"

// lib
#include <QFrustrationSharedGUI/global.h>

// Qt
#include <QtCore/QDebug>


ServerTab::ServerTab(QWidget *parent)
    : QWidget(parent),
    ui(new Ui::ServerTab)
{

    m_client = 0;

    ui->setupUi(this);
    connect(ui->joinGameButton, SIGNAL(clicked()), this, SLOT(joinGameClicked()));
    connect(ui->createButton, SIGNAL(clicked()), this, SLOT(createGameClicked()));
    connect(ui->gameList, SIGNAL(itemSelectionChanged()), this, SLOT(gameSelectionChanged()));

    ui->joinGameButton->setIcon(FRUST_ICON("network-connect"));
    ui->createButton->setIcon(FRUST_ICON("tab-new"));

}


ServerTab::~ServerTab()
{

    if (m_client) {
        m_client->disconnect(this);
        m_client->disconnectFromServer();
    }
    delete ui;

}


bool ServerTab::isQuickGameServer() const
{

    return !ui->createButton->isEnabled();

}


void ServerTab::setServer(QFC::Client *server)
{

    m_client = server;
    connect(server, SIGNAL(gotGameList(QHash<int,QString>)), this, SLOT(gotGameList(QHash<int,QString>)));
    connect(server, SIGNAL(gameOpen(int,QString)), this, SLOT(gameOpen(int,QString)));
    connect(server, SIGNAL(gameClosed(int,QString)), this, SLOT(gameClosed(int,QString)));
    connect(server, SIGNAL(disconnected(QFC::Client*)), this, SLOT(clientDisconnected()));
    server->requestGameList();

}


void ServerTab::startQuickGame(const QF::Layout &layout)
{

    ui->joinGameButton->setDisabled(true);
    ui->createButton->setDisabled(true);

    m_client->createGame(tr("Quick Game"), QString(), layout);

}


QList<QWidget*> ServerTab::gameTabs() const
{

    return m_gameTabs;

}


void ServerTab::addGameTab(QWidget *tab)
{

    m_gameTabs.append(tab);

}


void ServerTab::removeGameTab(QWidget *tab)
{

    m_gameTabs.removeAll(tab);

}


void ServerTab::gotGameList(const QHash<int, QString> &list)
{

    ui->gameList->clear();
    QHashIterator<int, QString> it(list);
    while (it.hasNext()) {
        it.next();

        QListWidgetItem *item = new QListWidgetItem;
        item->setText(it.value());
        item->setData(Qt::UserRole+1, it.key());

        ui->gameList->addItem(item);
    }

}


void ServerTab::gameOpen(const int &gameID, const QString &name)
{

    QListWidgetItem *item = new QListWidgetItem;
    item->setText(name);
    item->setData(Qt::UserRole+1, gameID);

    ui->gameList->addItem(item);

}


void ServerTab::gameClosed(const int &gameID, const QString &name)
{

    Q_UNUSED(name);
    for (int i = 0; i < ui->gameList->count(); i++) {
        QListWidgetItem *item = ui->gameList->item(i);

        if (item->data(Qt::UserRole+1).toInt() == gameID) {
            delete ui->gameList->takeItem(i);
            break;
        }
    }

}


void ServerTab::joinGameClicked()
{

    const QList<QListWidgetItem*> items = ui->gameList->selectedItems();
    if (items.isEmpty()) {
        return;
    }

    const int gameID = items.at(0)->data(Qt::UserRole+1).toInt();
    KeyDialog *dialog = new KeyDialog(gameID, this);
    connect(dialog, SIGNAL(finished(KeyDialog*)), this, SLOT(keyDialogFinished(KeyDialog*)));

    dialog->show();
    dialog->raise();

}


void ServerTab::createGameClicked()
{

    CreateGameDialog *dialog = new CreateGameDialog(this);
    connect(dialog, SIGNAL(finished(QString,QString,QF::Layout)), this,
            SLOT(createGameDialogFinished(QString,QString,QF::Layout)));

    dialog->show();
    dialog->raise();

}


void ServerTab::createGameDialogFinished(const QString &name, const QString &key, const QF::Layout &layout)
{

    m_client->createGame(name, key, layout);

}


void ServerTab::keyDialogFinished(KeyDialog *self)
{

    if (self->result() != QDialog::Accepted) {
        return;
    }
    m_client->joinGame(self->gameID(), self->key());

}


void ServerTab::clientDisconnected()
{

    emit disconnected(this);

}


void ServerTab::gameSelectionChanged()
{

    const QList<QListWidgetItem*> items = ui->gameList->selectedItems();
    ui->joinGameButton->setDisabled(items.isEmpty());

}
