/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "stonewidget.h"
#include "theme.h"
#include "config.h"

// Qt
#include <QtGui/QPainter>
#include <QtGui/QGraphicsSceneMouseEvent>
#include <QtGui/QIcon>
#include <QtGui/QPixmap>
#include <QtCore/QDebug>
#include <QtGui/QGraphicsScene>
#include <QtCore/QPropertyAnimation>
#include <QtGui/QGraphicsDropShadowEffect>
#include <QtGui/QGraphicsView>
#include <QtSvg/QSvgRenderer>
#include <QtCore/QSequentialAnimationGroup>


StoneWidget::StoneWidget(QF::Stone *s, const QColor &color, const bool &isPlayerStone, QGraphicsWidget *parent)
    : QGraphicsWidget(parent),
    m_stone(s),
    m_color(color),
    m_pressed(false)
{

    m_move = -1;

    m_isPlayerStone = isPlayerStone;

    m_scaleAnimation = new QPropertyAnimation(this);
    m_scaleAnimation->setTargetObject(this);
    m_scaleAnimation->setPropertyName("scale");
    m_scaleAnimation->setEasingCurve(QEasingCurve(QEasingCurve::Linear));

    m_moveGroup = new QSequentialAnimationGroup;
    connect(m_moveGroup, SIGNAL(finished()), this, SLOT(moveFinished()));

    connect(s, SIGNAL(moved(int,int)), this, SLOT(moved(int,int)));
    setAcceptHoverEvents(true);

    setZValue(5);

    setCacheMode(QGraphicsItem::DeviceCoordinateCache);

}



StoneWidget::~StoneWidget()
{

    delete m_scaleAnimation;
    delete m_moveGroup;

}


QF::Stone *StoneWidget::stone() const
{

    return m_stone;

}


int StoneWidget::possibleMove() const
{

    return m_move;

}


bool StoneWidget::isDown() const
{

    return m_pressed;

}


bool StoneWidget::isPlayerStone() const
{

    return m_isPlayerStone;

}


void StoneWidget::setPossibleMove(const int &move)
{

    m_move = move;

}


void StoneWidget::setDown(const bool &down)
{

    if (down == m_pressed) {
        return;
    }
    m_pressed = down;

    if (m_scaleAnimation->state() == QPropertyAnimation::Running) {
        m_scaleAnimation->stop();
    }

    m_scaleAnimation->setDuration(Config::self()->animationSpeed());
    m_scaleAnimation->setStartValue(scale());

    if (isDown()) {
        m_scaleAnimation->setEndValue(1.2);
    } else {
        m_scaleAnimation->setEndValue(1.0);
    }
    setTransformOriginPoint(rect().center());
    m_scaleAnimation->start();

}


void StoneWidget::stopMove()
{

    if (m_moveGroup->state() == QSequentialAnimationGroup::Running) {
        m_moveGroup->stop();
    }
    m_moveGroup->clear();

}


void StoneWidget::animateMove(const QList<QRectF> &fields)
{

    if (fields.isEmpty()) {
        return;
    }

    stopMove();

    QList<QRectF> list = fields;
    if (list.at(0) == geometry()) {
        list.removeAt(0);
    }

    int duration = (Config::self()->animationSpeed()/list.size())+(list.size()*15);

    QRectF start = geometry();
    foreach (const QRectF &rect, list) {
        QPropertyAnimation *ani = new QPropertyAnimation(this);
        ani->setEasingCurve(QEasingCurve::Linear);
        ani->setDuration(duration);
        ani->setPropertyName("geometry");
        ani->setTargetObject(this);
        ani->setStartValue(start);
        ani->setEndValue(rect);

        m_moveGroup->addAnimation(ani);
        start = rect;
    }

    setZValue(zValue()+1);
    m_moveGroup->start();

}


void StoneWidget::moved(const int &from, const int &to)
{

    emit stoneMoved(this, from, to);

}


void StoneWidget::moveFinished()
{

    setZValue(zValue()-1);
    m_moveGroup->clear();

}


void StoneWidget::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{

    Q_UNUSED(option);
    Q_UNUSED(widget);

    Theme::currentTheme()->renderStone(stone()->playerId(), painter, rect());

}


void StoneWidget::hoverEnterEvent(QGraphicsSceneHoverEvent *event)
{

    event->accept();
    if (isPlayerStone()) {
        setCursor(Qt::PointingHandCursor);
    }

}


void StoneWidget::hoverLeaveEvent(QGraphicsSceneHoverEvent *event)
{

    event->accept();
    if (isPlayerStone()) {
        setCursor(QCursor());
    }

}


void StoneWidget::mousePressEvent(QGraphicsSceneMouseEvent *event)
{

    event->accept();
    emit clicked(this);

}


void StoneWidget::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{

    event->accept();

}


void StoneWidget::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{

    event->accept();


}


