/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// own
#include "theme.h"
#include "desktopfile.h"

// lib
#include <QFrustrationSharedGUI/global.h>

// Qt
#include <QtCore/QDir>
#include <QtCore/QDebug>
#include <QtGui/QApplication>
#include <QtGui/QPainter>
#include <QtSvg/QSvgRenderer>
#include <QtCore/QSettings>

Theme::~Theme()
{

    if (m_renderer) {
        delete m_renderer;
    }

}


class ThemeSingleton
{
public:
    Theme self;
};


FRUSTRATION_SINGLETON(ThemeSingleton, themeSelf)


Theme *Theme::currentTheme()
{

    return &themeSelf->self;

}


QColor Theme::colorForRole(const Theme::ColorRole &role) const
{

    return m_currentTheme.colors.value(role);

}


QList<Theme::ThemeData> Theme::availableThemes() const
{

    QList<ThemeData> list;

    QString path = FrustGlobal::instance()->findFrustDir("themes");
    QDir dir(path);
    foreach (const QString desktopFile, dir.entryList(QStringList() << "*.desktop", QDir::Files)) {
        DesktopFile desktop(dir.absoluteFilePath(desktopFile));
        if (!desktop.init()) {
            qDebug() << "Error:" << desktop.error();
            continue;
        }

        ThemeData data;
        data.name = desktop.value("FrustTheme", "Name");
        data.description = desktop.value("FrustTheme", "Description");
        data.version = desktop.value("FrustTheme", "Version");
        data.email = desktop.value("FrustTheme", "Email");
        data.author = desktop.value("FrustTheme", "Author");
        data.file = dir.absoluteFilePath(desktop.value("FrustTheme", "FileName"));
        data.preview = dir.absoluteFilePath(desktop.value("FrustTheme", "Preview"));

        for (int i = 0; i < 6; i++) {
            data.colors.insert((Theme::ColorRole) i, QVariant(desktop.value("FrustTheme", QString("Player%1Color").arg(i))).value<QColor>());
        }
        list << data;
    }
    return list;

}


QString Theme::currentThemeName() const
{

    return m_currentTheme.name;

}


void Theme::setCurrentTheme(const QString &theme)
{

    ThemeData defaultTheme;
    const QList<ThemeData> list = availableThemes();
    foreach (const ThemeData &data, list) {
        if (data.name == theme) {
            m_currentTheme = data;
        }
        if (data.name == QString("Frusti")) {
            defaultTheme = data;
        }
    }

    if (m_renderer) {
        delete m_renderer;
    }

    Q_ASSERT(!defaultTheme.file.isEmpty());
    m_currentTheme = m_currentTheme.file.isEmpty() ? defaultTheme : m_currentTheme;

    m_renderer = new QSvgRenderer(m_currentTheme.file, this);
    emit themeChanged();

}


void Theme::renderBackground(QPainter *painter, const QRectF &rect)
{

    if (!rect.isValid()) {
        return;
    }

    m_renderer->render(painter, "Background", rect);

}



void Theme::renderStone(const int &id, QPainter *painter, const QRectF &rect)
{

    if (!rect.isValid()) {
        return;
    }

    QString key("p%1Stone");
    m_renderer->render(painter, key.arg(id), rect);

}


void Theme::renderField(QPainter *painter, const QRectF &rect, const int &player, const QF::Layout::FieldType &type)
{

    if (!rect.isValid()) {
        return;
    }

    QString key("p%1%2");
    key = key.arg(player);
    switch (type) {
    case QF::Layout::HomeType: key = key.arg("Home"); break;
    case QF::Layout::StartType: key = key.arg("Start"); break;
    case QF::Layout::TargetType: key = key.arg("Goal"); break; // FIXME
    case QF::Layout::WayType: key = "WayField"; break;
    case QF::Layout::LastType: key = "WayField"; break;
    default: return;
    }

    m_renderer->render(painter, key, rect);

}


qreal Theme::getMargin(const Theme::Margin &margin, const QString &element, const QRectF &rect) const
{

    QString key = QString("%1Margin%2").arg(element);
    switch (margin) {
    case LeftMargin: key = key.arg("Left"); break;
    case TopMargin: key = key.arg("Top"); break;
    case RightMargin: key = key.arg("Right"); break;
    case BottomMargin: key = key.arg("Bottom"); break;
    }

    if (!m_renderer->elementExists(key)) {
        qDebug() << "Element not found:" << key;
        return 0;
    }

    QRectF elementRect = m_renderer->boundsOnElement(key);
    const QSizeF size = rect.size();
    const QSizeF naturalSize = m_renderer->boundsOnElement(element).size();
    const qreal dx =  size.width()/naturalSize.width();
    qreal dy = size.height()/naturalSize.height();
    elementRect = QRectF(0, 0, elementRect.width()*dx, elementRect.height()*dy);

    switch (margin) {
    case LeftMargin:
    case RightMargin: return elementRect.width();
    case TopMargin:
    case BottomMargin: return elementRect.height();
    }

    return 0;

}


void Theme::getBoardMargins(qreal *left, qreal *top, qreal *right, qreal *bottom, const QRectF &rect)
{

    *left = getMargin(LeftMargin, "Board", rect);
    *top = getMargin(TopMargin, "Board", rect);
    *right = getMargin(RightMargin, "Board", rect);
    *bottom = getMargin(BottomMargin, "Board", rect);

}


void Theme::renderBoard(QPainter *painter, const QRectF &rect)
{

    if (!rect.isValid()) {
        return;
    }

    m_renderer->render(painter, "Board", rect);

}


void Theme::renderDice(QPainter *painter, const int &num, const QRectF &rect)
{

    if (!rect.isValid()) {
        return;
    }

    qreal left = getMargin(LeftMargin, "Dice", rect);
    qreal top = getMargin(TopMargin, "Dice", rect);
    qreal right = getMargin(RightMargin, "Dice", rect);
    qreal bottom = getMargin(BottomMargin, "Dice", rect);
    QRectF cRect = rect.adjusted(left, top, -right, -bottom);

    const qreal size = cRect.height()/4;
    QRectF dotRect = QRectF(0, 0, size, size);
    QPointF center = rect.center();

    QList<QRectF> dots;
    switch (num) {
    case 1: dotRect.moveCenter(center); dots.append(dotRect); break;
    case 2: {
            QRectF dot1(dotRect);
            dot1.moveCenter(center);
            dot1.moveLeft(cRect.left());

            QRectF dot2(dotRect);
            dot2.moveCenter(center);
            dot2.moveRight(cRect.right());

            dots << dot1 << dot2;
            break;
        }
    case 3: {
            QRectF dot1(dotRect);
            dot1.moveTopLeft(cRect.topLeft());

            QRectF dot3(dot1);
            dot3.moveBottomRight(cRect.bottomRight());

            QRectF dot2(dotRect);
            dot2.moveCenter(dot1.united(dot3).center());

            dots << dot1 << dot2 << dot3;
            break;
        }
    case 5: {
            QRectF dot5(dotRect);
            dot5.moveCenter(center);

            dots << dot5;
        }
    case 4: {
            QRectF dot1(dotRect);
            dot1.moveTopLeft(cRect.topLeft());

            QRectF dot2(dotRect);
            dot2.moveTopRight(cRect.topRight());

            QRectF dot3(dotRect);
            dot3.moveBottomLeft(cRect.bottomLeft());

            QRectF dot4(dotRect);
            dot4.moveBottomRight(cRect.bottomRight());

            dots << dot1 << dot2 << dot3 << dot4;
            break;
        }
    case 6: {
            QRectF dot1(dotRect);
            dot1.moveTopLeft(cRect.topLeft());

            QRectF dot2(dotRect);
            dot2.moveTopRight(cRect.topRight());

            QRectF dot3(dotRect);
            dot3.moveBottomLeft(cRect.bottomLeft());

            QRectF dot4(dotRect);
            dot4.moveBottomRight(cRect.bottomRight());

            QRectF dot5(dotRect);
            dot5.moveCenter(dot1.united(dot2).center());

            QRectF dot6(dotRect);
            dot6.moveCenter(dot3.united(dot4).center());

            dots << dot1 << dot2 << dot3 << dot4 << dot5 << dot6;
            break;
        }
    default: break;
    }

    m_renderer->render(painter, "Dice", rect);
    for (int i = 0; i < dots.size(); i++) {
        m_renderer->render(painter, "DiceDot", dots.at(i));
    }

}


Theme::Theme(QObject *parent)
    : QObject(parent)
{

    m_renderer = 0;
    setCurrentTheme("Frusti");

}


