/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "view.h"
#include "fieldwidget.h"
#include "stonewidget.h"
#include "dicewidget.h"
#include "theme.h"
#include "config.h"

// Qt
#include <QtCore/QDebug>
#include <QtGui/QPainter>
#include <QtGui/QPixmap>
#include <QtCore/QTimer>
#ifdef QT_QTOPENGL_FOUND
    #include <QtOpenGL/QGLWidget>
#endif


#define FIELDSIZE fieldSize()
#define FIELDSPACING 4.0
#define BOARDMARGIN Theme::currentTheme()->boardMargins(boardRect())
View::View(QWidget *parent)
    : QGraphicsView(parent)
{

    m_fieldSize = 4;

    m_resizeTimer = new QTimer(this);
    m_resizeTimer->setInterval(350);
    m_resizeTimer->setSingleShot(true);
    connect(m_resizeTimer, SIGNAL(timeout()), this, SLOT(delayedResize()));

    connect(this, SIGNAL(diceClicked()), this, SLOT(resetWidgetStates()));

    setScene(new QGraphicsScene(this));

    connect(Config::self(), SIGNAL(configChanged()), this, SLOT(configChanged()));
    configChanged();


}


View::~View()
{

    delete m_resizeTimer;

}


int View::currentZoom() const
{

    return transform().m11()*100;

}


void View::setLayout(const QF::Layout &layout)
{

    qDebug() << "setLayout:" << layout.rows() << layout.columns();

    setRenderHints(QPainter::Antialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform);
    setViewportUpdateMode(QGraphicsView::MinimalViewportUpdate);
    setCacheMode(QGraphicsView::CacheNone); // do it yourself
    setFrameStyle(QFrame::NoFrame);

    m_layout = layout;

    DiceWidget *dice = new DiceWidget;
    connect(dice, SIGNAL(clicked()), this, SIGNAL(diceClicked()));
    scene()->addItem(dice);
    m_dice = dice;

    QHashIterator<int, QF::Layout::Field> it(m_layout.getLayout());
    while (it.hasNext()) {
        it.next();

        FieldWidget *widget = new FieldWidget(QColor(), it.key(), 0);
        widget->setPlayerID(it.value().player);
        widget->setFieldType(it.value().type);
        connect(widget, SIGNAL(clicked(FieldWidget*)), this, SLOT(fieldClicked(FieldWidget*)));

        m_fields.insert(it.key(), widget);

        scene()->addItem(widget);
    }

    delayedResize(); // update sceneRect
    reLayout();

}


void View::addStone(QF::Stone *stone)
{

    const int size = FIELDSIZE;

    StoneWidget *widget = new StoneWidget(stone, QColor(), m_id == stone->playerId(), 0);
    m_stones.append(widget);
    connect(widget, SIGNAL(clicked(StoneWidget*)), this, SLOT(stoneClicked(StoneWidget*)));
    connect(widget, SIGNAL(stoneMoved(StoneWidget*,int,int)), this, SLOT(stoneMoved(StoneWidget*,int,int)));

    scene()->addItem(widget);

    FieldWidget *field = getField(stone->pos());
    Q_ASSERT(field);
    if (field) {
        QRectF rect(0, 0, size, size);
        rect.moveCenter(field->geometry().center());
        widget->setGeometry(rect);
    }

}


void View::setTurn(const QHash<int, int> &turn)
{

    for (int i = 0; i < items().size(); i++) {
        StoneWidget *widget = dynamic_cast<StoneWidget*>(items().at(i));
        if (!widget) {
            continue;
        }
        widget->setPossibleMove(-1);
    }

    if (turn.isEmpty()) {
        return;
    }

    qDebug() << "setTurn:" << turn;
    const QList<int> from = turn.keys();
    for (int i = 0; i < from.size(); i++) {
        const int &f = from.at(i);
        StoneWidget *widget = getStone(f);
        const int to = turn.value(f);
        Q_ASSERT(widget);

        qDebug() << "setMove:" << to;
        widget->setPossibleMove(to);
    }

}


void View::setRoll(const int &roll)
{

    m_dice->roll(roll);

}


void View::setPlayerList(const QHash<int, QString> list)
{

    Q_UNUSED(list);

}


void View::setCurrentPlayer(const int &playerID)
{

    if (playerID == m_id) { // reset
        foreach (StoneWidget *stone, m_stones) {
            stone->setPossibleMove(-1);
        }
    }

}


void View::setSelf(const int &id)
{

    m_id = id;

}


void View::enableOpenGL(const bool &enable)
{

#ifdef QT_QTOPENGL_FOUND
    if (enable) {
        setViewport(new QGLWidget(QGLFormat(QGL::SampleBuffers)));
    } else {
        setViewport(new QWidget(this));
    }
#else
    Q_UNUSED(enable);
#endif

}


FieldWidget *View::getField(const int &id) const
{

    return m_fields.value(id);

}


StoneWidget *View::getStone(const int &id) const
{

    foreach (StoneWidget *widget, m_stones) {
        if (widget->stone()->pos() == id) {
            return widget;
        }
    }
    return 0;

}


int View::playerId(const QString &nick) const
{

    Q_ASSERT(!m_players.isEmpty());
    qDebug() << "id for:" << nick << "players:" << m_players << "id:" << m_players.indexOf(nick);
    return m_players.indexOf(nick);

}


void View::moveRect(const View::Direction &direction, QRectF *rect, const int &count, const int &fieldSize)
{

    const int s = (FIELDSPACING+fieldSize)*count;
    switch (direction) {
    case View::Left: rect->moveLeft(rect->left()-s); break;
    case View::Right: rect->moveRight(rect->right()+s); break;
    case View::Up: rect->moveTop(rect->top()-s); break;
    case View::Down: rect->moveBottom(rect->bottom()+s); break;
    }

}


void View::resetWidgetStates(FieldWidget *skipField, StoneWidget *skipStone)
{

    QHashIterator<int, FieldWidget*> it(m_fields);
    while (it.hasNext()) {
        it.next();

        if (skipField && skipField == it.value()) {
            continue;
        }

        it.value()->setHighlighted(false);
    }
    for (int i = 0; i < m_stones.size(); i++) {
        StoneWidget *widget = m_stones.at(i);

        if (skipStone && skipStone == widget) {
            continue;
        }

        widget->setDown(false);
    }

}


QRectF View::topLeftFieldRect() const
{

    qreal left, top, right, bottom;
    Theme::currentTheme()->getBoardMargins(&left, &top, &right, &bottom, boardRect());

    QRectF r(0, 0, FIELDSIZE, FIELDSIZE);
    r.moveTopLeft(boardRect().topLeft()+QPointF(left+FIELDSPACING, top+FIELDSPACING));
    return r;

}


StoneWidget *View::getCurrentStone() const
{

    for (int i = 0; i < m_stones.size(); i++) {
        StoneWidget *widget = m_stones.at(i);

        if (widget->isDown()) {
            return widget;
        }
    }
    return 0;

}


QRectF View::boardRect() const
{

    if (!viewport()) {
        return QRectF();
    }

    int rows = m_layout.rows();
    int cols = m_layout.columns();

    QRectF min = QRectF(0, 0, qMin(sceneRect().height(), sceneRect().width()), qMin(sceneRect().width(), sceneRect().height()));

    qreal left, top, right, bottom;
    Theme::currentTheme()->getBoardMargins(&left, &top, &right, &bottom, min);

    int w = cols*(FIELDSIZE+FIELDSPACING);
    int h = rows*(FIELDSIZE+FIELDSPACING);
    h += FIELDSPACING+top+bottom;
    w += FIELDSPACING+left+right;

    QRectF rect;
    rect.setWidth(w);
    rect.setHeight(h);
    rect.moveCenter(sceneRect().center());

    return rect;

}


int View::fieldSize() const
{

    return m_fieldSize;

}


void View::reLayout()
{

    QRectF dicePos = topLeftFieldRect();
    dicePos.moveCenter(sceneRect().center());
    m_dice->setGeometry(dicePos);

    QHash<int, QF::Layout::Field> f = m_layout.getLayout();
    const QRectF topLeft = topLeftFieldRect();
    const int size = fieldSize();
    foreach (FieldWidget *widget, m_fields) {
        QRectF geo = topLeft;
        moveRect(Right, &geo, f.value(widget->id()).column, size);
        moveRect(Down, &geo, f.value(widget->id()).row, size);

        widget->setGeometry(geo);
    }

    // stones
    for (int i = 0; i < m_stones.size(); i++) {
        StoneWidget *widget = m_stones.at(i);

        FieldWidget *field = getField(widget->stone()->pos());
        Q_ASSERT(field);
        if (field) {
            QRectF rect = topLeft;
            rect.moveCenter(field->geometry().center());
            widget->stopMove();
            widget->setGeometry(rect);
        }
    }

}


void View::stoneMoved(StoneWidget *self, const int &from, const int &to)
{

    resetWidgetStates();

    QList<int> indexList = m_layout.getWay(self->stone()->playerId(), from, to);
    QList<QRectF> rectList;
    foreach (const int &index, indexList) {
        FieldWidget *field = getField(index);
        Q_ASSERT(field);

        rectList << field->geometry();
    }
    self->animateMove(rectList);

}



void View::fieldClicked(FieldWidget *field)
{

    Q_ASSERT(field->stone);
    if (!field->isHighlighted() || field->stone->stone()->pos() == field->id()) {
        return;
    }
    emit sendTurn(field->stone->stone()->pos(), field->id());

}


void View::stoneClicked(StoneWidget *self)
{

    if (sender()) {
        // mouse navi
        if (Config::self()->value("View-SingleClick", false)) {
            resetWidgetStates();
            FieldWidget *field = getField(self->possibleMove());
            if (field) {
                field->stone = self;
                field->setHighlighted(true);
                fieldClicked(field);
            }
        } else {
            // click on a stone of an opponent?
            StoneWidget *current = getCurrentStone();
            if (!self->isPlayerStone() && current) {
                FieldWidget *field = getField(self->stone()->pos()); // field under stone

                Q_ASSERT(field);
                if (current->possibleMove() == field->id()) {
                    fieldClicked(field);
                }
            } else {
                if (self->possibleMove() != -1) {
                    FieldWidget *widget = getField(self->possibleMove());
                    Q_ASSERT(widget);

                    resetWidgetStates(widget, self);

                    widget->stone = self;
                    self->setDown(true);
                    widget->setHighlighted(true);
                } else {
                    resetWidgetStates();
                }
            }
        }
    } else {
        // keyboard navi
        if (self->possibleMove() != -1) {
            FieldWidget *widget = getField(self->possibleMove());
            Q_ASSERT(widget);

            resetWidgetStates(widget, self);

            widget->setHighlighted(true);
            widget->stone = self;

            self->setDown(true);
        }
    }

}


void View::delayedResize()
{

    if (!viewport()) {
        return;
    }

    QSizeF size = viewport()->contentsRect().size();
    if (size.width() < m_layout.columns()*20) {
        size.rwidth() = m_layout.columns()*20;
    }
    if (size.height() < m_layout.rows()*20) {
        size.rheight() = m_layout.rows()*20;
    }

    setSceneRect(QRectF(QPointF(0, 0), size));

    int w = sceneRect().width();
    int h = sceneRect().height();
    const int min = qMin(w, h);

    QRectF board = QRectF(0, 0, min, min);

    qreal left, top, right, bottom;
    Theme::currentTheme()->getBoardMargins(&left, &top, &right, &bottom, board);

    h = ((board.height()-top)-bottom)-(FIELDSPACING*2);
    w = ((board.width()-left)-right)-(FIELDSPACING*2);

    const int rows = m_layout.rows();
    const int cols = m_layout.columns();

    m_fieldSize = h/rows;
    if (w/cols < m_fieldSize) {
        m_fieldSize = w/cols;
    }

    m_fieldSize -= FIELDSPACING;

    viewport()->update();  // avoid artifacts
    reLayout();

}


void View::configChanged()
{

    enableOpenGL(Config::self()->openGL());

}


void View::drawBackground(QPainter *painter, const QRectF &rect)
{

    if (sceneRect().height() < m_layout.rows()*20 ||
            sceneRect().width() < m_layout.columns()*20) {
        return;
    }

    painter->fillRect(rect, backgroundBrush());
    if (!m_background.isNull() && m_background.size() == sceneRect().size().toSize()) {
        QRect r = rect.adjusted(-8, -8, 16, 16).toRect().normalized(); // avoid artifacts
        painter->drawPixmap(r, m_background, r);
        return;
    }

    m_background = QPixmap(sceneRect().size().toSize());
    m_background.fill(Qt::transparent);

    QPainter p(&m_background);
    p.setRenderHints(painter->renderHints());

    Theme::currentTheme()->renderBackground(&p, sceneRect());
    Theme::currentTheme()->renderBoard(&p, boardRect());
    p.end();

    painter->drawPixmap(sceneRect().topLeft(), m_background);

/*
    qDebug() << "rows/cols:" << m_layout.rows() << m_layout.columns();

    // grid
    QRectF fr = topLeftFieldRect();
    fr.moveTop(fr.top()-(FIELDSIZE));
    for (int r = 0; r < m_layout.rows(); r++) {
        fr.moveLeft(topLeftFieldRect().left());
        fr.moveBottom(topLeftFieldRect().bottom()+((FIELDSIZE+FIELDSPACING)*r));
        for (int c = 0; c < m_layout.columns(); c++) {
            painter->drawRect(fr);
            fr.moveLeft(fr.left()+FIELDSPACING+FIELDSIZE);
        }
    }
*/

}


void View::resizeEvent(QResizeEvent *event)
{

    QGraphicsView::resizeEvent(event);
    if (!viewport()) {
        return;
    }

    if (m_resizeTimer->isActive()) {
        m_resizeTimer->stop();
    }
    m_resizeTimer->start();

}


void View::keyPressEvent(QKeyEvent *event)
{

    event->accept();

}


void View::keyReleaseEvent(QKeyEvent *event)
{

    event->accept();

    const Qt::Key nextKey = Config::self()->key(Config::NextStoneKeyType);
    const Qt::Key rollKey = Config::self()->key(Config::RollKeyType);
    const Qt::Key sentKey = Config::self()->key(Config::SentKeyType);

    if (event->key() == rollKey) {
        emit diceClicked();
    } else if (event->key() == nextKey) {
        QList<StoneWidget*> playerStones;
        for (int i = 0; i < m_stones.size(); i++) {
            StoneWidget *w = m_stones.at(i);
            if (w->isPlayerStone() && w->possibleMove() != -1) {
                playerStones.append(w);
            }
        }

        if (playerStones.isEmpty()) {
            return;
        }

        StoneWidget *current = getCurrentStone();
        int index = current ? playerStones.indexOf(current) : -1;
        ++index;

        if (index >= playerStones.size()) {
            index = 0;
        }

        current = playerStones.at(index);
        Q_ASSERT(current);
        current->setDown(true);
        stoneClicked(current);
    } else if (event->key() == sentKey) { // send turn
        StoneWidget *current = getCurrentStone();
        if (current) {
            FieldWidget *field = getField(current->possibleMove());
            if (field) {
                fieldClicked(field);
            }
        }
    }

}


void View::scaleBy(qreal scaleFactor)
{

    qreal minScale = 0.2;
    qreal maxScale = 5.0;

    qreal curScaleFactor = transform().m11();
    if (((curScaleFactor == minScale) && (scaleFactor < 1.0)) || ((curScaleFactor == maxScale) && (scaleFactor > 1.0))) {
        return;
    }
    qreal sc = scaleFactor;

    if ((curScaleFactor * sc < minScale)&&(sc < 1.0))  {
        sc = minScale / curScaleFactor;
    } else if ((curScaleFactor * sc > maxScale) && (sc > 1.0)) {
        sc = maxScale / curScaleFactor;
    }
    scale(sc, sc);

}


void View::setZoom(int percentZoom)
{

    if (percentZoom < 10) {
        return;
    }

    percentZoom = (qRound((qreal)percentZoom/10.0))*10.0;

    if (percentZoom == 100) {
        resetMatrix();
        return;
    }

    qreal targetScale = (qreal)percentZoom / 100.0;
    qreal scaleFactor = targetScale / transform().m11();
    scaleBy(scaleFactor);

}
