/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/



// own
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "propertyeditor.h"

// lib
#include <QFrustrationSharedGUI/aboutdialog.h>
#include <QFrustrationSharedGUI/global.h>
#include <QFrustrationSharedGUI/errordialog.h>
#include <QFrustrationSharedGUI/questiondialog.h>

// Qt
#include <QtGui/QMessageBox>
#include <QtGui/QFileDialog>
#include <QtCore/QDebug>
#include <QtGui/QToolBar>
#include <QtGui/QAction>
#include <QtGui/QApplication>


MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent),
    ui(new Ui::MainWindow)
{

    ui->setupUi(this);
    ui->rowsSpin->setValue(ui->view->rows());
    ui->columnsSpin->setValue(ui->view->columns());
    ui->playersSpin->setValue(ui->view->players());
    ui->stonesSpin->setValue(ui->view->stones());

    connect(ui->addButton, SIGNAL(clicked()), ui->view, SLOT(addField()));
    connect(ui->view, SIGNAL(itemSelected(FieldWidget*)), ui->editor, SLOT(setField(FieldWidget*)));
    connect(ui->view, SIGNAL(editModeChanged(MapMakerView::EditMode)), this, SLOT(editModeChanged(MapMakerView::EditMode)));

    QMenu *fileMenu = new QMenu(this);
    fileMenu->setTitle(tr("File"));

    QAction *saveAction = getAction("SaveAction");
    saveAction->setText(tr("Save Layout").append("..."));
    saveAction->setIcon(FRUST_ICON("document-save"));
    connect(saveAction, SIGNAL(triggered()), this, SLOT(saveClicked()));

    QAction *loadAction = getAction("LoadAction");
    loadAction->setText(tr("Load Layout").append("..."));
    loadAction->setIcon(FRUST_ICON("document-open"));
    connect(loadAction, SIGNAL(triggered()), this, SLOT(loadClicked()));


    QMenu *helpMenu = new QMenu(this);
    helpMenu->setTitle(tr("Help"));

    QAction *aboutAction = new QAction(this);
    aboutAction->setText(tr("About %1 ...").arg(qApp->applicationName()));
    aboutAction->setIcon(windowIcon());
    connect(aboutAction, SIGNAL(triggered()), this, SLOT(showAbout()));

    QAction *aboutQtAction = new QAction(this);
    aboutQtAction->setText(tr("About %1 ...").arg("Qt"));
    aboutQtAction->setIcon(FRUST_ICON("dialog-information"));
    connect(aboutQtAction, SIGNAL(triggered()), qApp, SLOT(aboutQt()));

    fileMenu->addAction(saveAction);
    fileMenu->addAction(loadAction);

    helpMenu->addAction(aboutAction);
    helpMenu->addAction(aboutQtAction);

    menuBar()->addMenu(fileMenu);
    menuBar()->addMenu(helpMenu);


    QToolBar *bar = new QToolBar(tr("Toolbar"), this);

    QAction *wayAction = getAction("WayModeAction");
    wayAction->setText(tr("Way"));
    wayAction->setIcon(FRUST_ICON("draw-line"));
    wayAction->setCheckable(true);
    connect(wayAction, SIGNAL(triggered(bool)), this, SLOT(wayModetriggered(bool)));

    QAction *addAction = getAction("AddModeAction");
    addAction->setText(tr("Add"));
    addAction->setIcon(FRUST_ICON("list-add"));
    addAction->setCheckable(true);
    connect(addAction, SIGNAL(triggered(bool)), this, SLOT(addModetriggered(bool)));

    QAction *removeAction = getAction("RemoveModeAction");
    removeAction->setText(tr("Remove"));
    removeAction->setIcon(FRUST_ICON("list-remove"));
    removeAction->setCheckable(true);
    connect(removeAction, SIGNAL(triggered(bool)), this, SLOT(removeModetriggered(bool)));

    QAction *showCodeAction = getAction("ShowCodeAction");
    showCodeAction->setText(tr("Show Code"));
    showCodeAction->setIcon(FRUST_ICON("document-edit"));
    connect(showCodeAction, SIGNAL(triggered()), ui->view, SLOT(showCode()));

    bar->addAction(addAction);
    bar->addAction(removeAction);
    bar->addAction(wayAction);
    bar->addSeparator();
    bar->addAction(showCodeAction);

    addToolBar(Qt::TopToolBarArea, bar);

}


MainWindow::~MainWindow()
{

    delete ui;

}


QAction *MainWindow::getAction(const QString &name)
{

    foreach (QAction *act, actions()) {
        if (act->objectName() == name) {
            return act;
        }
    }

    QAction *action = new QAction(this);
    insertAction(0, action);
    action->setObjectName(name);
    return action;

}


void MainWindow::saveClicked()
{

    if (!ui->view->isValid()) {
        QuestionDialog *dialog = QuestionDialog::showQuestion(this,
                                                              tr("Warning"),
                                                              tr("The current layout is invalid, save anyway?"));
        connect(dialog, SIGNAL(yesClicked()), this, SLOT(getSaveFile()));
    } else {
        getSaveFile();
    }

}


void MainWindow::getSaveFile()
{

    QFileDialog *dialog = new QFileDialog(this);
    dialog->setAttribute(Qt::WA_DeleteOnClose, true);
    connect(dialog, SIGNAL(fileSelected(QString)), this, SLOT(saveLayout(QString)));
    dialog->setConfirmOverwrite(true);
    dialog->setAcceptMode(QFileDialog::AcceptSave);
    dialog->show();
    dialog->raise();

}


void MainWindow::saveLayout(const QString &file)
{

    QFile f(file);
    if ((f.exists() && !f.remove()) || !f.open(QIODevice::WriteOnly)) {
        ErrorDialog::showError(this, tr("Error"), f.errorString());
    } else {
        QByteArray data;
        ui->view->save(&data);
        f.write(data);
        f.close();
    }

}


void MainWindow::loadClicked()
{

    QFileDialog *dialog = new QFileDialog(this);
    dialog->setAttribute(Qt::WA_DeleteOnClose, true);
    connect(dialog, SIGNAL(fileSelected(QString)), this, SLOT(loadLayout(QString)));
    dialog->setAcceptMode(QFileDialog::AcceptOpen);
    dialog->setNameFilter(tr("QFrustration Layout (*.frust)"));
    dialog->show();
    dialog->raise();

}


void MainWindow::loadLayout(const QString &file)
{

    QFile f(file);
    if (!f.open(QIODevice::ReadOnly)) {
        ErrorDialog::showError(this, tr("Error"), f.errorString());
    } else {
        const QByteArray data = f.readAll();
        ui->view->load(data);
    }

}


void MainWindow::wayModetriggered(const bool &checked)
{

    const MapMakerView::EditMode mode = checked ? MapMakerView::WayMode : MapMakerView::NormalMode;
    ui->view->setEditMode(mode);

}


void MainWindow::addModetriggered(const bool &checked)
{

    const MapMakerView::EditMode mode = checked ? MapMakerView::AddMode : MapMakerView::NormalMode;
    ui->view->setEditMode(mode);
}


void MainWindow::removeModetriggered(const bool &checked)
{

    const MapMakerView::EditMode mode = checked ? MapMakerView::RemoveMode : MapMakerView::NormalMode;
    ui->view->setEditMode(mode);

}


void MainWindow::editModeChanged(const MapMakerView::EditMode &mode)
{

    QAction *currentAction = 0;
    switch (mode) {
    case MapMakerView::WayMode: currentAction = getAction("WayModeAction"); break;
    case MapMakerView::AddMode: currentAction = getAction("AddModeAction"); break;
    case MapMakerView::RemoveMode: currentAction = getAction("RemoveModeAction"); break;
    case MapMakerView::NormalMode: break;
    }

    QList<QAction*> actions;
    actions << getAction("WayModeAction") <<
               getAction("AddModeAction") <<
               getAction("RemoveModeAction");
    if (currentAction) {
        actions.removeAll(currentAction);
        currentAction->setChecked(true);
    }

    foreach (QAction *action, actions) {
        action->setChecked(false);
    }

}


void MainWindow::showAbout()
{

    QHash<QString, QString> devs;
    devs.insert("Kai Dombrowe", "just89@gmx.de");
    AboutDialog::showAbout(this,
                           devs,
                           QUrl("https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=CGJG8DG4H72W2"));

}



