/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// own
#include "mapmakerview.h"
#include "fieldwidget.h"
#include "buildwaydialog.h"

// lib
#include <QFrustrationLib/layout.h>

// Qt
#include <QtGui/QPainter>
#include <QtGui/QPixmap>
#include <QtCore/QDebug>
#include <QtCore/QSettings>
#include <QtGui/QMessageBox>
#include <QtGui/QTextBrowser>
#include <QtCore/QFile>


#define FIELDSIZE 40.0
#define FIELDSPACING 12.0
MapMakerView::MapMakerView(QWidget *parent)
    : QGraphicsView(parent)
{

    m_hoverRect = QRectF(-1, -1, -1, -1);
    m_stones = 1;
    m_players = 2;
    m_rows = 4;
    m_columns = 4;

    m_mode = NormalMode;

    setScene(new QGraphicsScene(this));
    setFields(11, 11);

}


MapMakerView::~MapMakerView()
{



}


bool MapMakerView::isValid() const
{

    return buildLayout().isValid();

}


void MapMakerView::save(QByteArray *data)
{

    QF::Layout layout = buildLayout();
    layout.save(data);

}


void MapMakerView::load(const QByteArray &data)
{

    scene()->clear();

    QF::Layout layout;
    layout.load(data);

    if (!layout.isValid()) {
        qWarning() << "invalid layout";
    }

    setFields(layout.rows(), layout.columns());
    setPlayers(layout.players());
    setStones(layout.stones());

    QHashIterator<int, QF::Layout::Field> it(layout.getLayout());
    while (it.hasNext()) {
        it.next();

        const QF::Layout::Field f = it.value();

        FieldWidget *widget = createField();
        widget->setIndex(it.key());
        widget->setFieldType(f.type);
        widget->setPlayerID(f.player);
        widget->setGeometry(getRect(f.row, f.column));
    }

}


void MapMakerView::showCode()
{

    QTextBrowser *b = new QTextBrowser(0);
    b->setAttribute(Qt::WA_DeleteOnClose, true);
    b->setWindowTitle(tr("Code Generator"));

    b->append(QString("setPlayers(%1);").arg(players()));
    b->append(QString("setStones(%1);").arg(stones()));
    b->append(QString("setRows(%1);").arg(rows()));
    b->append(QString("setColumns(%1);").arg(columns()));

    for (int f = 0; f < items().size(); f++) {
        FieldWidget *widget = static_cast<FieldWidget*>(items().at(f));

        QString typeStr = QString("QF::Layout::");
        switch(widget->fieldType()) {
        case QF::Layout::WayType: typeStr.append("WayType"); break;
        case QF::Layout::HomeType: typeStr.append("HomeType"); break;
        case QF::Layout::StartType: typeStr.append("StartType"); break;
        case QF::Layout::LastType: typeStr.append("LastType"); break;
        case QF::Layout::TargetType: typeStr.append("TargetType"); break;
        case QF::Layout::InvalidType: typeStr.append("InvalidType"); break;
        }

        const QString line = QString("addField(%1, %2, %3, %4, %5);").arg(typeStr).arg(widget->index()).arg(
                widget->playerID()).arg(getRow(widget)).arg(getColumn(widget));
        b->append(line);
    }
    b->show();

}


void MapMakerView::setFields(const int &rows, const int &columns)
{

    if (columns < 3 || columns > 300|| rows < 3 || rows > 300) {
        qWarning() << "Invalid rows/columns:" << rows << columns;
        return;
    }

    m_rows = rows;
    m_columns = columns;

    emit rowsChanged(m_rows);
    emit columnsChanged(m_columns);

    //setSceneRect(0, 0, (rows*FIELDSIZE)+((rows-1)*FIELDSPACING), (columns*FIELDSIZE)+((columns-1)*FIELDSPACING));
    QRectF rect;
    rect.setWidth((columns*FIELDSIZE)+((columns-1)*FIELDSPACING));
    rect.setHeight((rows*FIELDSIZE)+((rows-1)*FIELDSPACING));

    qDebug() << "width:" << rect.width();
    qDebug() << "height:" << rect.height();

    setSceneRect(rect);

}


void MapMakerView::setEditMode(const MapMakerView::EditMode &mode)
{

    if (mode != WayMode && m_mode == WayMode && !m_wayPoints.isEmpty()) {
        BuildWayDialog dialog(this);
        if (dialog.exec() == QDialog::Accepted) {
            int index = dialog.startIndex();
            foreach (const QPointF &pos, m_wayPoints) {
                FieldWidget *widget  = static_cast<FieldWidget*>(itemAt(mapFromScene(pos)));
                Q_ASSERT(widget);
                widget->setIndex(index);
                widget->setFieldType(dialog.fieldType());
                widget->setPlayerID(dialog.playerID());
                index++;
            }
        }
        m_wayPoints.clear();
        viewport()->update();
    }

    m_mode = mode;
    emit editModeChanged(mode);

}


int MapMakerView::players() const
{

    return m_players;

}


int MapMakerView::stones() const
{

    return m_stones;

}


int MapMakerView::rows() const
{

    return m_rows;

}


int MapMakerView::columns() const
{

    return m_columns;

}


void MapMakerView::setPlayers(int players)
{

    m_players = players;

    emit playersChanged(m_players);

}


void MapMakerView::setStones(int stones)
{

    m_stones = stones;
    emit stonesChanged(m_stones);

}


void MapMakerView::setRows(int rows)
{

    m_rows = rows;
    setFields(m_rows, m_columns);

    emit rowsChanged(m_rows);

}


void MapMakerView::setColumns(int columns)
{

    m_columns = columns;
    setFields(m_rows, m_columns);

    emit columnsChanged(m_columns);

}


void MapMakerView::addField()
{

    FieldWidget *widget = createField();
    widget->setGeometry(getRect(0, 0));

}


QList<QRectF> MapMakerView::fieldRects() const
{

    QList<QRectF> rects;
    QRectF rect(QPointF(0, 0), QSizeF(FIELDSIZE, FIELDSIZE));
    rect.moveTopLeft(sceneRect().topLeft());
    for (int r = 0; r < m_rows; r++) {
        if (r != 0) {
            rect.moveBottom(rect.bottom()+FIELDSIZE+FIELDSPACING);
        }
        rect.moveLeft(sceneRect().left());
        for (int c = 0;  c < m_columns; c++) {
            if (c != 0) {
                rect.moveRight(rect.right()+FIELDSIZE+FIELDSPACING);
            }
            rects << rect;
        }
    }
    return rects;

}

void MapMakerView::test()
{
    foreach (QGraphicsItem *item, items()) {
        FieldWidget *field = static_cast<FieldWidget*>(item);
        qDebug() << "ROW:" << getRow(field);
        qDebug() << "COL:" << getColumn(field);
    }

}

int MapMakerView::getRow(FieldWidget *widget) const
{

    int row = widget->geometry().y()/(FIELDSIZE+FIELDSPACING);
    return row;

}


int MapMakerView::getColumn(FieldWidget *widget) const
{

    int col = widget->geometry().x()/(FIELDSIZE+FIELDSPACING);
    return col;
}


QRectF MapMakerView::getRect(int row, int column) const
{

    QRectF rect;
    rect.setWidth(FIELDSIZE);
    rect.setHeight(FIELDSIZE);
    rect.moveTopLeft(QPointF(column*(FIELDSIZE+FIELDSPACING), row*(FIELDSIZE+FIELDSPACING)));
    return rect;

}


FieldWidget *MapMakerView::createField()
{

    FieldWidget *widget = new FieldWidget;
    connect(widget, SIGNAL(moved(FieldWidget*,QPointF)), this, SLOT(moved(FieldWidget*,QPointF)));
    connect(widget, SIGNAL(moveDone(FieldWidget*,QPointF)), this, SLOT(moveDone(FieldWidget*,QPointF)));
    connect(widget, SIGNAL(pressed(FieldWidget*)), this, SIGNAL(itemSelected(FieldWidget*)));

    scene()->addItem(widget);

    return widget;

}


QF::Layout MapMakerView::buildLayout() const
{

    QF::Layout layout;
    layout.beginLayout(rows(), columns(), players(), stones());

    QList<QGraphicsItem*> itemList = items();
    for (int f = 0; f < itemList.size(); f++) {
        FieldWidget *widget = static_cast<FieldWidget*>(itemList.at(f));

        layout.addField(widget->fieldType(), widget->index(), widget->playerID(), getRow(widget), getColumn(widget));
    }
    layout.endLayout();
    return layout;

}



void MapMakerView::setFieldPos(QGraphicsWidget *widget, int row, int column)
{


    QRectF rect(QPointF(0, 0), QSizeF(FIELDSIZE, FIELDSIZE));
    rect.moveTopLeft(sceneRect().topLeft());
    for (int r = 0; r < m_rows; r++) {
        if (r != 0) {
            rect.moveBottom(rect.bottom()+FIELDSIZE+FIELDSPACING);
            rect.moveLeft(sceneRect().left());
        }
        for (int c = 0;  c < m_columns; c++) {
            if (c != 0) {
                rect.moveRight(rect.right()+FIELDSIZE+FIELDSPACING);
            }
            if (c == column && r == row) {
                qDebug() << "setGeometry:" << rect << row << column;
                widget->setGeometry(rect);
                return;
            }
        }
    }

}


void MapMakerView::moved(FieldWidget *self, const QPointF &to)
{

    if (m_mode ==WayMode) {
        return;
    }

    QRectF geo = self->geometry();
    geo.moveCenter(to);
    self->setGeometry(geo);


}


void MapMakerView::moveDone(FieldWidget *self, const QPointF &to)
{

    if (m_mode == WayMode) {
        m_wayPoints.append(self->geometry().center());
        viewport()->update();
        return;
    }

    moved(self, to);

    bool found = false;
    foreach (const QRectF &r, fieldRects()) {
        if (r.contains(self->geometry().center())) {
            m_hoverRect = r;
            found = true;
            break;
        }
    }

    if (found) {
        self->setGeometry(m_hoverRect);
    } else {
        self->setGeometry(fieldRects().first());
    }
    m_hoverRect = QRectF(-1, -1, -1, -1);

}


void MapMakerView::drawBackground(QPainter *painter, const QRectF &rect)
{

    painter->fillRect(rect, Qt::black);
    painter->setPen(Qt::red);
    foreach (const QRectF &r, fieldRects()) {
        if (m_hoverRect == r) {
            painter->fillRect(r, Qt::red);
        }
        painter->drawRect(r);
    }

}


void MapMakerView::drawForeground(QPainter *painter, const QRectF &rect)
{

    Q_UNUSED(rect);

    bool first = true;
    QPainterPath path;
    foreach (const QPointF pos, m_wayPoints) {
        if (first) {
            path.moveTo(pos);
            first = false;
        } else {
            path.lineTo(pos);
        }
    }

    painter->setRenderHints(QPainter::Antialiasing);
    painter->setPen(QPen(Qt::green, 6));
    painter->drawPath(path);

    painter->setBrush(QBrush(QColor(Qt::green).darker(), Qt::SolidPattern));
    foreach (const QPointF pos, m_wayPoints) {
        painter->drawEllipse(pos, 10, 10);
    }

}


void MapMakerView::mousePressEvent(QMouseEvent *event)
{

    if (m_mode == AddMode || m_mode == RemoveMode) {
        event->accept();
    } else {
        QGraphicsView::mousePressEvent(event);
    }

}


void MapMakerView::mouseReleaseEvent(QMouseEvent *event)
{

    if (m_mode == AddMode) {
        event->accept();

        const QPointF pos = mapToScene(event->pos());
        foreach (const QRectF &r, fieldRects()) {
            if (r.contains(pos)) {
                FieldWidget *widget = createField();
                widget->setGeometry(r);
                break;
            }
        }
    } else if (m_mode == RemoveMode) {
        QGraphicsItem *item = itemAt(event->pos());
        if (item) {
            scene()->removeItem(item);
            delete item;
        }
    } else {
        QGraphicsView::mouseReleaseEvent(event);
    }

}









