/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "client.h"
#include "client_p.h"
#include "server.h"

// lib
#include <QFrustrationLib/protocol.h>

// Qt
#include <QtCore/QDebug>
#include <QtCore/QTimerEvent>


namespace QFS {


ClientPrivate::ClientPrivate(Client *parent, Server *s)
    : QTcpSocket(parent),
    server(s),
    q(parent)
{

    connect(this, SIGNAL(disconnected()), this, SLOT(clientDisconnected()));
    connect(this, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(socketEerror(QAbstractSocket::SocketError)));
    connect(this, SIGNAL(connected()), this, SLOT(clientConnected()));
    connect(this, SIGNAL(readyRead()), this, SLOT(incomingData()));
    connect(this, SIGNAL(bytesWritten(qint64)), q, SIGNAL(dataSent(qint64)));

}


ClientPrivate::~ClientPrivate()
{



}


void ClientPrivate::ping()
{

  //  qDebug() << "write ping!";
    write(QF::Protocol::ping());
    flush();

}


void ClientPrivate::pong()
{

   // qDebug() << "got pong";
    gotPong = true;

}


void ClientPrivate::clientDisconnected()
{

    qDebug() << "client disconnected...";
    emit q->disconnected(q);

}


void ClientPrivate::socketEerror(const QAbstractSocket::SocketError &error)
{

    qDebug() << "Error:" << error;

}


void ClientPrivate::clientConnected()
{

   // qDebug() << "CONNECTED!";

}


void ClientPrivate::incomingData()
{

    emit q->receivingData(bytesAvailable());
    m_buffer.append(readAll());

    QByteArray blob;
    while (QF::Protocol::getBlobFromBuffer(&m_buffer, &blob)) {
        QStringList args;
        QF::Protocol::BlobType type;

        if (!QF::Protocol::parseBlob(blob, &type, &args)) {
            qFatal("PARSE ERROR");
        }

        if (type != QF::Protocol::PongType && nick.isEmpty() && type != QF::Protocol::SetNickType) {
            write(QF::Protocol::sendError(QF::Protocol::NoNickSetErrorType));
            return;
        }

        switch (type) {
        case QF::Protocol::PongType: pong(); break;
        case QF::Protocol::SetNickType: {
                QString nick;
                if (QF::Protocol::parseRequestNick(&args, &nick)) {
                    q->setNick(nick);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        case QF::Protocol::RollType: {
                int gameID;
                if (QF::Protocol::parseRequestRoll(&args, &gameID)) {
                    emit q->rollRequested(q, gameID);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        case QF::Protocol::CreateGameType: {
                QString name;
                QString key;
                QF::Layout layout;
                if (QF::Protocol::parseCreateGame(&args, &name, &key, &layout)) {
                    emit q->gameCreateRequested(q, name, key, layout);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        case QF::Protocol::JoinGameTye: {
                int gameID;
                QString key;

                if (QF::Protocol::parseJoinGame(&args, &gameID, &key)) {
                    emit q->joinGameRequested(q, gameID, key);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        case QF::Protocol::StartGameRequestType: {
                int gameID;

                if (QF::Protocol::parseStartGameRequest(&args, &gameID)) {
                    emit q->startGameRequested(q, gameID);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        case QF::Protocol::GameListType: {
                emit q->gameListRequested(q);
                break;
            }
        case QF::Protocol::TurnType: {
                int from;
                int to;
                int gameID;
                if (QF::Protocol::parseRequestTurn(&args, &from, &to, &gameID)) {
                    emit q->turnRequested(q, from, to, gameID);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        case QF::Protocol::GameChatType: {
                QString text;
                QString sender;
                int gameID;

                if (QF::Protocol::parseGameChat(&args, &text, &sender, &gameID)) {
                    emit q->newGameChatText(q, text, sender, gameID);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        case QF::Protocol::GameLeftType: {
                int gameID;

                if (QF::Protocol::parseLeave(&args, &gameID)) {
                    emit q->leaveRequested(q, gameID);
                } else {
                    write(QF::Protocol::sendError(QF::Protocol::InvalidDataErrorType));
                }
                break;
            }
        default: qDebug() << "INVALID BLOB:" << blob;
        }
    }
   // qDebug() << "BUFFER LEFT:" << m_buffer;

}


void ClientPrivate::timerEvent(QTimerEvent *event)
{

    if (event->timerId() == pingTimer) {
        if (!gotPong) {
            qDebug() << "gotPong = false";
            close();
        } else {
            gotPong = false;
            ping();
        }
    } else {
        QTcpSocket::timerEvent(event);
    }

}



Client::Client(QObject *parent, Server *server)
    : QObject(parent),
    d(new ClientPrivate(this, server))
{



}


Client::~Client()
{

}


QString Client::nick() const
{

    return d->nick;

}


bool Client::init(const int &socket)
{

    if (!d->setSocketDescriptor(socket)) {
        return false;
    }

    d->gotPong = true;
    d->pingTimer = d->startTimer(60*1000); // 1 minute
    d->ping();

    return true;

}


void Client::setNick(const QString &nick)
{


    if (!nick.isEmpty() && d->server->isNickValid(nick)) {
      //  qDebug() << "setNick:" << nick;
        d->nick = nick;
        d->write(QF::Protocol::setNick(nick));

        emit connected(this);
    } else {
        //qDebug() << "setNick failed" << nick;
        d->write(QF::Protocol::sendError(QF::Protocol::NickAlreadyInUseErrorType));
    }

}


void Client::send(const QByteArray &blob)
{

    d->write(blob);

}


} // namespace QFS
