/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "game.h"
#include "client.h"

// lib
#include <QFrustrationLib/stone.h>
#include <QFrustrationLib/layout.h>
#include <QFrustrationLib/protocol.h>

// Qt
#include <QtCore/QDebug>
#include <QtCore/QDateTime>


namespace QFS {


Game::Game(QFS::Client *owner, const int &gameId, const QString &name, const QString &key,
           const QF::Layout &layout, QObject *parent)
    : QObject(parent),
    m_owner(owner),
    m_id(gameId)
{

    m_maxPlayers = 4;
    m_clients.insert(owner, 0);
    m_currentPlayer = -1;
    m_rollCount = 1;
    m_playerHasRolled = false;
    m_started = false;
    m_name = name;
    m_key = key;

    m_layout = layout;

 /*   qDebug() << "LAYOUT:" << layout.getLayout().keys();
    QHash<int, QF::Layout::Field> fs = layout.getLayout();
    QHashIterator<int, QF::Layout::Field> it(fs);
    while (it.hasNext()) {
        it.next();

        qDebug() << "Index:" << it.key() << "Row:" << it.value().row << "Col:" << it.value().column;
    }*/
    Q_ASSERT(layout.isValid());

}


Game::~Game()
{



}


int Game::gameId() const
{

    return m_id;

}


QFS::Client *Game::owner() const
{

    return m_owner;

}


QList<QFS::Client*> Game::clients() const
{

    return m_clients.keys();

}


QString Game::name() const
{

    return m_name;

}


QString Game::key() const
{

    return m_key;

}


QF::Layout Game::layout() const
{

    return m_layout;

}


void Game::startGame(QFS::Client *starter)
{

    if (m_started) {
        starter->send(QF::Protocol::sendError(QF::Protocol::GameAlreadyStartedErrorType));
        return;
    }
    if (starter != owner()) {
        starter->send(QF::Protocol::sendError(QF::Protocol::NotGameOwnerErrorType));
        return;
    }

    if (m_clients.size() < 2) {
        starter->send(QF::Protocol::sendError(QF::Protocol::NotEnoughPlayersErrorType));
        return;
    }
    m_started = true;

    emit gameStarted(this);

    Q_ASSERT(m_layout.isValid());

    sendToClients(QF::Protocol::sendGameStarted(gameId()));

    QHashIterator<Client*, int> it(m_clients);
    while (it.hasNext()) {
        it.next();

        const int playerID = it.value();
        const QList<int> homeIndexList = m_layout.homes(playerID);
        for (int stoneID = 0; stoneID < homeIndexList.size(); stoneID++) {
            QF::Stone *stone = new QF::Stone(this, playerID, stoneID, gameId());
            stone->setPos(homeIndexList.at(stoneID));

            m_stones.append(stone);
        }
    }

    for (int i = 0; i < m_stones.size(); i++) {
        connect(m_stones.at(i), SIGNAL(moved(int,int)), this, SLOT(stoneMoved(int,int)));
        sendToClients(QF::Protocol::sendStoneAdded(m_stones.at(i)));
    }
    nextTurn();

}


void Game::addPlayer(QFS::Client *client)
{

    if (m_started) {
        client->send(QF::Protocol::sendError(QF::Protocol::GameAlreadyStartedErrorType));
        return;
    }

    const int max = m_layout.players();
    if (m_clients.size()+1 > max) {
        client->send(QF::Protocol::sendError(QF::Protocol::TooManyPlayersErrorType));
        return;
    }

    int id = 0;
    QList<int> ids = m_clients.values();
    while (ids.contains(id)) {
        id++;
    }
    const QString player = client->nick();
    m_clients.insert(client, id);

    QHash<int, QString> list;
    QHashIterator<QFS::Client*, int> it(m_clients);
    while (it.hasNext()) {
        it.next();

        list.insert(it.value(), it.key()->nick());
    }

    sendToClients(QF::Protocol::sendGameJoined(player, id, gameId(), name(), list, layout()));

}


void Game::removePlayer(QFS::Client *client)
{

    client->disconnect(this);

    const int id = m_clients.take(client);
    const QString player = client->nick();
    sendToClients(QF::Protocol::sendGameLeft(player, id, gameId()));

    if (!m_started && client != owner()) {
        return;
    }

    if (m_clients.size() < 2 || client == owner()) {
        emit gameClosed(this);
    } else {
        if (m_currentPlayer == id) {
            nextTurn();
        }
    }

}


void Game::setMaximumPlayers(const int &max)
{

    m_maxPlayers = max;

}


bool Game::hasStoneOnField(const int &player) const
{

    const QList<QF::Stone*> stones = stonesFromPlayer(player);
    const QList<int> home = m_layout.homes(player);
    const QList<int> Target = m_layout.targets(player);
    for (int i = 0; i < stones.size(); i++) {
        const QF::Stone *stone = stones.at(i);

        if (!home.contains(stone->pos()) && !Target.contains(stone->pos())) {
            return true;
        }
    }
    return false;

}


void Game::setTurn(QFS::Client *client, const int &from, const int &to)
{

  //  qDebug() << "setTurn:" << from << to;

    if (!m_clients.contains(client)) {
        qDebug() << "Client not in game";
        return;
    }

    if (!m_started) {
        client->send(QF::Protocol::sendError(QF::Protocol::GameNotStartedErrorType));
        return;
    } else if (m_clients.value(client) != m_currentPlayer) {
        client->send(QF::Protocol::sendError(QF::Protocol::NotYourTurnErrorType));
        return;
    } else if (!m_playerHasRolled) {
        client->send(QF::Protocol::sendError(QF::Protocol::PlayerHasNotRolledErrorType));
        return;
    }

   // qDebug() << "Turn empty?" << m_currentTurn.isEmpty();
    if (m_currentTurn.isEmpty()) {
        nextTurn();
        return;
    }

    if (!m_currentTurn.values(from).contains(to)) {
        // invalid turn
        qDebug() << "invalid turn" << "from:" << from << "to" << to << "possible turn:" << m_currentTurn;
        client->send(QF::Protocol::sendError(QF::Protocol::InvalidTurnErrorType));
        Q_ASSERT(false);
        return;
    }

    bool kicked = false;
    QF::Stone *stone = stoneOnIndex(from);
    QF::Stone *onDest = stoneOnIndex(to);
    if (onDest) {
        kicked = true;
        const int home = getFreeHome(onDest);
        Q_ASSERT(home != -1);
        onDest->setPos(home);
    }
    stone->setPos(to);


    // winner?
    bool winner = true;
    QList<QF::Stone*> playerStones = stonesFromPlayer(m_currentPlayer);
    const QList<int> targets = m_layout.targets(m_currentPlayer);
    for (int i = 0; i < playerStones.size(); i++) {
        QF::Stone *stone = playerStones.at(i);

        if (!targets.contains(stone->pos())) {
            winner = false;
            break;
        }
    }

    if (winner) {
        sendToClients(QF::Protocol::sendGameWinner(gameId(), m_currentPlayer));
        return;
    }

    // missed kick
    if (!kicked) {
        QHashIterator<int, int> it(m_currentTurn);
        while (it.hasNext()) {
            it.next();

            QF::Stone *stoneTo = stoneOnIndex(it.value());
            QF::Stone *stoneFrom;
            if (it.key() == from) {
                stoneFrom = stoneOnIndex(to);
            } else {
                stoneFrom = stoneOnIndex(it.key());
            }
            if (stoneTo && stoneFrom && stoneTo->playerId() != m_currentPlayer) {
            //    qDebug() << "player:" << m_currentPlayer << "missed a kick";
                const int home = getFreeHome(stoneFrom);
                Q_ASSERT(home != -1);
                stoneFrom->setPos(home);
            }
        }
    }
    nextTurn();

}


void Game::createRoll(QFS::Client *client)
{

    if (!m_clients.contains(client)) {
        qDebug() << "Client not in game";
        return;
    }

    if (!m_started) {
        client->send(QF::Protocol::sendError(QF::Protocol::GameNotStartedErrorType));
        return;
    }

    if (m_clients.value(client) != m_currentPlayer)  {
        client->send(QF::Protocol::sendError(QF::Protocol::NotYourTurnErrorType));
        return;
    }

    if (m_playerHasRolled) {
        client->send(QF::Protocol::sendError(QF::Protocol::PlayerHasAlreadyRolledErrorType));
        return;
    }

    m_currentRoll = roll();
    ++m_rollCount;
    m_playerHasRolled = true;

    sendToClients(QF::Protocol::sendRoll(m_currentRoll, gameId()));
    createTurn();

}


void Game::sendChat(const QString &sender, const QString &text)
{

    sendToClients(QF::Protocol::sendGameChat(text, sender, gameId()));

}


void Game::sendToClients(const QByteArray &blob)
{

    QHashIterator<Client*, int> it(m_clients);
    while (it.hasNext()) {
        it.next();

        it.key()->send(blob);
    }

}


void Game::nextTurn()
{

    if (m_currentPlayer != -1) {
        if (m_currentRoll == 6) {
            m_rollCount = 1;
            m_playerHasRolled = false;
            sendToClients(QF::Protocol::sendPlayerChanged(m_currentPlayer, gameId()));
            return;
        } else if (m_rollCount < 4) {
            if (!hasStoneOnField(m_currentPlayer) && !hasHoleInTarget(m_currentPlayer) && m_currentTurn.isEmpty()) {
                m_playerHasRolled = false;
                sendToClients(QF::Protocol::sendPlayerChanged(m_currentPlayer, gameId()));
                return;
            }
        }
    }

    if (m_currentPlayer == -1) {
        // first turn
        m_currentPlayer = (qrand() % m_clients.size()-1)+1;
    } else {
        // next turn
        int index = m_currentPlayer;
        index++;
        if (index >= m_clients.size()) {
            index = 0;
        }
        m_currentPlayer = index;
    }

    Q_ASSERT(m_clients.values().contains(m_currentPlayer));

    m_rollCount = 1;
    m_playerHasRolled = false;

    sendToClients(QF::Protocol::sendPlayerChanged(m_currentPlayer, gameId()));

}


int Game::roll() const
{

    return (qrand() % 6) + 1;

}


QList<QF::Stone*> Game::stonesInHome(const int &player) const
{

    QList<QF::Stone*> list;
    for (int i = 0; i < m_stones.size(); i++) {
        QF::Stone *stone = m_stones.at(i);
        if (stone->playerId() == player) {
            if (m_layout.homes(player).contains(stone->pos())) {
                list.append(stone);
            }
        }
    }
    return list;

}


QF::Stone *Game::stoneOnStart(const int &player) const
{

    for (int i = 0; i < m_stones.size(); i++) {
        QF::Stone *stone = m_stones.at(i);
        if (stone->playerId() == player) {
            if (m_layout.startField(player) == stone->pos()) {
                return stone;
            }
        }
    }
    return 0;

}


QList<QF::Stone*> Game::stonesFromPlayer(const int &player) const
{

    QList<QF::Stone*> list;
    for (int i = 0; i < m_stones.size(); i++) {
        QF::Stone *stone = m_stones.at(i);
        if (stone->playerId() == player) {
            list.append(stone);
        }
    }
    return list;

}


QF::Stone *Game::stoneOnIndex(const int &index) const
{

    for (int i = 0; i < m_stones.size(); i++) {
        QF::Stone *stone = m_stones.at(i);
        if (stone->pos() == index) {
            return stone;
        }
    }
    return 0;

}


bool Game::hasHoleInTarget(const int &player) const
{

    QList<QF::Stone*> pStones = stonesFromPlayer(player);
    QList<int> targets = m_layout.targets(player);

    QList<QF::Stone*> inTarget;
    for (int i = 0; i < pStones.size(); i++) {
        if (targets.contains(pStones.at(i)->pos())) {
            inTarget.append(pStones.at(i));
        }
    }

    bool lastTargetTaken = true; // 5
    while (!targets.isEmpty()) {
        const int Target = targets.takeLast();
        bool currentTargetTaken = false;
        for (int s = 0; s < inTarget.size(); s++) {
            QF::Stone *stone = inTarget.at(s);
            if (stone->pos() == Target) {
                currentTargetTaken = true;
                break;
            }
        }
        if (currentTargetTaken && !lastTargetTaken) {
            return true;
        } else {
            lastTargetTaken = currentTargetTaken;
        }
    }
    return false;

}


int Game::getFreeHome(QF::Stone *stone) const
{

    const QList<int> home = m_layout.homes(stone->playerId());
    for (int i = 0; i < home.size(); i++) {
        if (!stoneOnIndex(home.at(i))) {
            return home.at(i);
        }
    }
    return -1;

}


void Game::createTurn()
{

    QHash<int, int> turn;
    Client *client = m_clients.key(m_currentPlayer);
    Q_ASSERT(client);


    QList<QF::Stone*> playerStones = stonesFromPlayer(m_currentPlayer);
    QList<QF::Stone*> inHome = stonesInHome(m_currentPlayer);
    QF::Stone *onStart = stoneOnStart(m_currentPlayer);
    if (m_currentRoll == 6) {
        if (onStart && playerStones.contains(onStart)) {
            // nothing to do...
        } else {
            if (!inHome.isEmpty()) {
                for (int i = 0; i < inHome.size(); i++) {
                    turn.insert(inHome.at(i)->pos(), m_layout.startField(m_currentPlayer));
                }
                m_currentTurn = turn;
                client->send(QF::Protocol::sendTurn(turn, gameId()));
                return;
            }
        }
    }

    if (onStart && onStart->playerId() == m_currentPlayer && !inHome.isEmpty()) { // own stone on start
        const int index = m_layout.tryMove(m_currentPlayer, onStart->pos(), m_currentRoll);
        QF::Stone *stone = stoneOnIndex(index);
        if (index != -1 && (!stone || !playerStones.contains(stone))) { // possable?
            turn.insert(onStart->pos(), index);
            m_currentTurn = turn;
            client->send(QF::Protocol::sendTurn(turn, gameId()));
            return;
        }
    }

    for (int i = 0; i < inHome.size(); i++) {
        playerStones.removeAll(inHome.at(i));
    }

    for (int i = 0; i < playerStones.size(); i++) {
        QF::Stone *stone = playerStones.at(i);
        int index = m_layout.tryMove(m_currentPlayer, stone->pos(), m_currentRoll);

        // do not jump over other stones
        QList<int> targets = m_layout.targets(m_currentPlayer);
        if (targets.contains(index)) {
            const int from = targets.indexOf(stone->pos());
            for (int s = 0; s < playerStones.size(); s++) {
                QF::Stone *stone2 = playerStones.at(s);
                if (stone == stone2) {
                    continue;
                }
                if (targets.contains(stone2->pos()) && stone2->pos() < index && targets.indexOf(stone2->pos()) > from) {
                    index = -1;
                    break;
                }
            }

        }

      //  qDebug() << "try to move from:" << stone->pos() << "with roll" << m_currentRoll << "success:" << index;
        if (index != -1) {
            QF::Stone *onIndex = stoneOnIndex(index);
            if (!onIndex || onIndex->playerId() != m_currentPlayer) {
                turn.insert(stone->pos(), index);
            }
        }
    }

    m_currentTurn = turn;
   // qDebug() << "turn created:" << turn;

    if (turn.isEmpty()) {
        nextTurn();
    } else {
        client->send(QF::Protocol::sendTurn(turn, gameId()));
    }

}


void Game::stoneMoved(const int &from, const int &to)
{

    sendToClients(QF::Protocol::sendStoneMoved(from, to, gameId()));

}


void Game::rollRequested(QFS::Client *client)
{

    createRoll(client);

}


void Game::gotTurn(QFS::Client *self, const int &from, const int &to)
{

    setTurn(self, from, to);

}


} // namespace QFS
