/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// own
#include "protocol.h"
#include "stone.h"

// Qt
#include <QtXml/QXmlStreamWriter>
#include <QtXml/QXmlStreamReader>
#include <QtCore/QDebug>


namespace QF {


bool Protocol::getBlobFromBuffer(QByteArray *buffer, QByteArray *blob)
{

    const QByteArray start = "<BlobSize";
    const bool startsWith = buffer->startsWith(start);
    if (!startsWith && buffer->size() >= start.size()) {
        qDebug() << "got invalid data";
        if (blob->contains(start)) {
            blob->remove(0, blob->indexOf(start, 0)); // remove trash
            return getBlobFromBuffer(buffer, blob);
        }
        if (buffer->size() > 1024*2) { // remove trash
            buffer->clear();
        }
        return false;
    } else if (!startsWith) {
        return false;
    }


    QRegExp sizeRX("(<BlobSize \\d{1,}>\n)");

    QString text = *buffer;
    if (!text.contains(sizeRX)) {
        qDebug() << "Size not found....";
        return false;
    }

    if (sizeRX.indexIn(text) == -1) {
        qDebug() << "indexIn() failed...";
        return false;
    }

    QString sizeString = sizeRX.cap(0);
    const int sizeStringSize = sizeString.size();
    sizeString.remove("<BlobSize");
    sizeString.remove('>');
    sizeString = sizeString.trimmed();

    bool ok = false;
    const int blobSize = sizeString.toInt(&ok);
    if (!ok) {
        qDebug() << "toInt() failed:" << sizeString;
        return false;
    }

    if (buffer->size() < blobSize+sizeStringSize) {
        qDebug() << "wrong size:" << buffer->size() << blobSize;
        return false;
    }

    text.remove(sizeRX);
    text.resize(blobSize);

    buffer->remove(0, text.size()+sizeRX.cap(0).size()); // remove blob from buffer

    blob->clear();
    blob->append(text);

    return true;

}


QByteArray Protocol::createBlob(const QF::Protocol::BlobType &type, const QStringList &args)
{

    QByteArray blob;
    QXmlStreamWriter stream(&blob);
    stream.setAutoFormatting(true);
    stream.writeStartDocument();

    stream.writeStartElement("Blob");
    stream.writeAttribute("Type", QString::number(type));

    stream.writeStartElement("Data");
    stream.writeAttribute("Args", QString::number(args.size()));
    for (int i = 0; i < args.size(); i++) {
        stream.writeAttribute(QString("Arg%1").arg(i), args.at(i));
    }
    stream.writeEndElement(); // Data
    stream.writeEndElement(); // Blob

    stream.writeEndDocument();

    blob.append("\n");
    blob.prepend(QString("<BlobSize %1>\n").arg(blob.size()).toLatin1());

    return blob;

}


bool Protocol::parseBlob(QByteArray &blob, QF::Protocol::BlobType *type, QStringList *args)
{

    QByteArray xml = blob;
    xml.remove(0, blob.indexOf('>', 0)+1); // <BlobSize XXXX>

    QXmlStreamReader reader(xml.trimmed());
    while (reader.readNextStartElement()) {
        if (reader.name() == QLatin1String("Blob")) {
            bool ok = false;
            const int typeInt = reader.attributes().value("Type").toString().toInt(&ok);

            if (!ok) {
                *type = InvalidType;
                return false;
            } else {
                *type = static_cast<BlobType>(typeInt);
            }
        } else if (reader.name() == QLatin1String("Data")) {
            bool ok = false;
            const int argCount = reader.attributes().value("Args").toString().toInt(&ok);
            if (ok) {
                for (int i = 0; i < argCount; i++) {
                    args->append(reader.attributes().value(QString("Arg%1").arg(i)).toString());
                }
            } else {
                return false;
            }
        } else {
            reader.skipCurrentElement();
        }
    }

    if (reader.hasError()) {
        qDebug() << "ERROR: " << reader.errorString();
        return false;
    }
    return true;

}


QByteArray Protocol::sendGameChat(const QString &text, const QString &sender, const int &gameID)
{

    return createBlob(GameChatType, QStringList() << text << sender << QString::number(gameID));

}


QByteArray Protocol::ping()
{

    return createBlob(PingType, QStringList());

}


QByteArray Protocol::setNick(const QString &nick)
{

    return createBlob(SetNickType, QStringList() << nick);

}


QByteArray Protocol::sendGameStarted(const QStringList &playerList, const QString &gameID)
{

    return createBlob(Protocol::GameStartType, QStringList() <<
                      QString::number(playerList.size()) <<
                      playerList <<
                      gameID);

}


QByteArray Protocol::sendStoneAdded(QF::Stone *stone)
{

    return createBlob(StoneAddedType, QStringList() <<
                      QString::number(stone->playerId()) <<
                      QString::number(stone->stoneId()) <<
                      QString::number(stone->pos()) <<
                      QString::number(stone->gameId()));

}


QByteArray Protocol::sendRoll(const int &value, const int &gameID)
{

    return createBlob(RollType, QStringList() << QString::number(value) << QString::number(gameID));

}


QByteArray Protocol::sendTurn(const QHash<int, int> &turn, const int &gameID)
{

    QStringList args;

    QList<int> keys = turn.keys();
    args << QString::number(keys.size());

    for (int i = 0; i < keys.size(); i++) {
        args << QString::number(keys.at(i));

        QList<int> values = turn.values(keys.at(i));

        args << QString::number(values.size());
        for (int v = 0; v < values.size(); v++) {
            args << QString::number(values.at(v));
        }
    }
    args << QString::number(gameID);
    return createBlob(TurnType, args);

}


QByteArray Protocol::sendStoneMoved(const int &from, const int &to, const int &gameID)
{

    return createBlob(StoneMovedType, QStringList() << QString::number(from) << QString::number(to) << QString::number(gameID));

}


QByteArray Protocol::sendGameLeft(const QString &player, const int &id, const int &gameID)
{

    return createBlob(GameLeftType, QStringList() << player << QString::number(id) << QString::number(gameID));

}


QByteArray Protocol::sendError(const QF::Protocol::Error &error)
{

    return createBlob(ErrorType, QStringList() << QString::number(error));

}


QByteArray Protocol::sendGameWinner(const int &gameID, const int &playerID)
{

    return createBlob(GameWinnerType, QStringList() << QString::number(gameID) << QString::number(playerID));

}


QByteArray Protocol::sendPlayerChanged(const int &playerID, const int &gameID)
{

    return createBlob(GamePlayerChangedType, QStringList() << QString::number(playerID) << QString::number(gameID));

}


QByteArray Protocol::sendGameStarted(const int &gameID)
{

    return createBlob(GameStartType, QStringList() << QString::number(gameID));

}


QByteArray Protocol::sendGameInvite(const QString &from, const int &gameID)
{

    return createBlob(GameInviteType, QStringList() << from << QString::number(gameID));

}


static bool readLayout(const QStringList *args, int index, QF::Layout *layout)
{

    bool ok = false;

    int players = args->at(index).toInt();
    index++;
    int stones = args->at(index).toInt();
    index++;
    int rows = args->at(index).toInt();
    index++;
    int columns = args->at(index).toInt();
    index++;

    layout->beginLayout(rows, columns, players, stones);

    int count = args->at(index).toInt(&ok);
    if (!ok) {
        return false;
    }
    index++;
    for (int i = 0; i < count; i++) {
        const int key = args->at(index).toInt(&ok);
        if (!ok) {
            return false;
        }
        index++;

        QF::Layout::Field field;
        field.type = (QF::Layout::FieldType) args->at(index).toInt(&ok);
        if (!ok) {
            return false;
        }
        index++;

        field.player = args->at(index).toInt(&ok);
        if (!ok) {
            return false;
        }
        index++;

        field.row = args->at(index).toInt(&ok);
        if (!ok) {
            return false;
        }
        index++;

        field.column = args->at(index).toInt(&ok);
        if (!ok) {
            return false;
        }
        index++;

        ok = layout->addField(field.type, key, field.player, field.row, field.column);
        if (!ok) {
            return false;
        }
    }
    layout->endLayout();
    return true;

}


static void writeLayout(QStringList *args, const QF::Layout *layout)
{

    *args << QString::number(layout->players());
    *args << QString::number(layout->stones());
    *args << QString::number(layout->rows());
    *args << QString::number(layout->columns());

    QHash<int, QF::Layout::Field> fields = layout->getLayout();
    *args << QString::number(fields.count());
    QHashIterator<int, QF::Layout::Field> it(fields);
    while (it.hasNext()) {
        it.next();

        *args << QString::number(it.key());

        *args << QString::number((int)it.value().type);
        *args << QString::number(it.value().player);
        *args << QString::number(it.value().row);
        *args << QString::number(it.value().column);
    }

}



QByteArray Protocol::sendGameCreated(const int &gameID, const QString &name, const QString &key, const QF::Layout &layout)
{

    QStringList args = QStringList() << QString::number(gameID) << name << key;
    writeLayout(&args, &layout);
    return createBlob(CreateGameType, args);

}


QByteArray Protocol::sendGameJoined(const QString &playerName,
                                    const int &playerID,
                                    const int &gameID,
                                    const QString &gameName,
                                    const QHash<int, QString> &playerList,
                                    const QF::Layout &layout)
{

    QStringList args;
    args << playerName;
    args << QString::number(playerID);
    args << QString::number(gameID);
    args << gameName;

    args << QString::number(playerList.size());
    QHashIterator<int, QString> it(playerList);
    while (it.hasNext()) {
        it.next();

        args << QString::number(it.key());
        args << it.value();
    }

    writeLayout(&args, &layout);

    return createBlob(GameJoinedType, args);

}


QByteArray Protocol::sendGameList(const QHash<int, QString> &list)
{

    QStringList args;
    args << QString::number(list.size());

    QHashIterator<int, QString> it(list);
    while (it.hasNext()) {
        it.next();

        args << QString::number(it.key());
        args << it.value();
    }

    return createBlob(GameListType, args);

}


QByteArray Protocol::sendGameOpen(const int &gameID, const QString &name)
{

    return createBlob(GameOpenType, QStringList() << QString::number(gameID) << name);

}


QByteArray Protocol::sendGameClosed(const int &gameID, const QString &name)
{

    return createBlob(GameClosedType, QStringList() << QString::number(gameID) << name);

}


QByteArray Protocol::pong()
{

    return createBlob(PongType, QStringList());

}


QByteArray Protocol::requestJoinGame(const int &gameID, const QString &key)
{

    return createBlob(JoinGameTye, QStringList() << QString::number(gameID) << key);

}


QByteArray Protocol::requestStart(const int &gameID)
{

    return createBlob(StartGameRequestType, QStringList() << QString::number(gameID));

}


QByteArray Protocol::requestNick(const QString &nick)
{

    return createBlob(SetNickType, QStringList() << nick);

}


QByteArray Protocol::requestRoll(const int &gameID)
{

    return createBlob(RollType, QStringList() << QString::number(gameID));

}


QByteArray Protocol::requestTurn(const int &from, const int &to, const int &gameID)
{

    return createBlob(TurnType, QStringList() << QString::number(from) << QString::number(to) << QString::number(gameID));

}


QByteArray Protocol::sendCreateGame(const QString &name, const QString &key, const QF::Layout &layout)
{

    QStringList args;
    args << name << key;
    writeLayout(&args, &layout);
    return createBlob(CreateGameType, args);

}


QByteArray Protocol::requestGameList()
{

    return createBlob(GameListType, QStringList());

}


QByteArray Protocol::requestLeave(const int &gameID)
{

    return createBlob(Protocol::GameLeftType, QStringList() << QString::number(gameID));

}



bool Protocol::parseServerSetsNick(const QStringList *args, QString *nick)
{

    if (args->size() > 1 || args->size() < 1) {
        return false;
    }

    *nick = args->at(0);
    return true;

}


bool Protocol::parseError(const QStringList *args, QF::Protocol::Error *error)
{

    if (args->size() > 1 || args->size() < 1) {
        return false;
    }

    bool ok = false;
    *error = (Error) args->at(0).toInt(&ok);
    return ok;

}


bool Protocol::parseGameStarted(const QStringList *args, int *gameID)
{

    if (args->size() > 1 || args->size() < 1) {
        return false;
    }
    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    return ok;

}


bool Protocol::parseCurrentPlayerChanged(const QStringList *args, int *playerID, int *gameID)
{

    if (args->size() > 2 || args->size() < 2) {
        return false;
    }
    bool ok = false;
    *playerID = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameID = args->at(1).toInt(&ok);
    return ok;

}


bool Protocol::parseStoneAdded(const QStringList *args, int *playerID, int *stoneID, int *stonePos, int *gameID)
{

    if (args->size() > 4 || args->size() < 4) {
        return false;
    }

    bool ok = false;
    *playerID = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }
    *stoneID = args->at(1).toInt(&ok);
    if (!ok) {
        return false;
    }
    *stonePos = args->at(2).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameID = args->at(3).toInt(&ok);
    return ok;

}


bool Protocol::parseRoll(const QStringList *args, int *value, int *gameID)
{

    if (args->size() > 2 || args->size() < 2) {
        return false;
    }

    bool ok = false;
    *value = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameID = args->at(1).toInt(&ok);
    return ok;

}


bool Protocol::parseTurn(const QStringList *args, QHash<int, int> *turn, int *gameID)
{

    if (args->size() < 1) {
        return false;
    }

    bool ok = false;

    int index = 0;
    const int keyCount = args->at(index).toInt(&ok);
    index++;

    if (!ok) {
        return false;
    }

    for (int i = 0; i < keyCount; i++) {
        if (args->size() <= index) {
            return false;
        }
        const int key = args->at(index).toInt(&ok);
        index++;

        if (!ok) {
            return false;
        }

        if (args->size() <= index) {
            return false;
        }

        const int valueCount = args->at(index).toInt(&ok);
        index++;

        if (!ok) {
            return false;
        }


        for (int v = 0; v < valueCount; v++) {
            if (args->size() <= index) {
                return false;
            }

            const int value = args->at(index).toInt(&ok);
            index++;

            if (!ok) {
                return false;
            }

            turn->insertMulti(key, value);
        }
    }

    if (args->size() <= index) {
        return false;
    }
    *gameID = args->at(index).toInt(&ok);
    return ok;

}


bool Protocol::parseStoneMoved(const QStringList *args, int *from, int *to, int *gameID)
{

    if (args->size() > 3 || args->size() < 3) {
        return false;
    }

    bool ok = false;
    *from = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }
    *to = args->at(1).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameID = args->at(2).toInt(&ok);
    return ok;

}


bool Protocol::parseGameCreated(const QStringList *args, int *gameID, QString *name, QString *key, QF::Layout *layout)
{

    if (args->size() < 4) {
        return false;
    }
    bool ok = false;

    *gameID = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }
    *name = args->at(1);
    *key = args->at(2);

    ok = readLayout(args, 3, layout);

    return ok;

}


bool Protocol::parseGameLeft(const QStringList *args, QString *nick, int *playerID, int *gameID)
{

    if (args->size() != 3) {
        return false;
    }

    *nick = args->at(0);

    bool ok = false;
    *playerID = args->at(1).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameID = args->at(2).toInt(&ok);
    return ok;

}


bool Protocol::parseGameList(const QStringList *args, QHash<int, QString> *gameList)
{

    if (args->size() < 1) {
        return false;
    }

    bool ok = false;
    const int size = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }

    if (args->size() != (size*2)+1) {
        return false;
    }

    int index = 1;
    for (int i = 0; i < size; i++) {
        const int id = args->at(index).toInt(&ok);
        if (!ok) {
            return false;
        }
        index++;
        const QString name = args->at(index);
        index++;

        gameList->insert(id, name);
    }

    return true;

}



bool Protocol::parseGameOpen(const QStringList *args, int *gameID, QString *name)
{

    if (args->size() != 2) {
        return false;
    }

    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    *name = args->at(1);

    return ok;

}


bool Protocol::parseGameClosed(const QStringList *args, int *gameID, QString *name)
{

    if (args->size() != 2) {
        return false;
    }

    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    *name = args->at(1);

    return ok;

}


bool Protocol::parseGameJoined(const QStringList *args,
                               QString *playerName,
                               int *playerID,
                               int *gameID,
                               QString *gameName,
                               QHash<int, QString> *playerList,
                               QF::Layout *layout)
{

    if (args->size() < 5) {
        return false;
    }

    *playerName = args->at(0);

    bool ok = false;
    *playerID = args->at(1).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameID = args->at(2).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameName = args->at(3);

    const int size = args->at(4).toInt(&ok);
    if (!ok) {
        return false;
    }

    int index = 5;
    for (int i = 0; i < size; i++) {
        if (args->size() < index) {
            return false;
        }

        const int id = args->at(index).toInt(&ok);
        if (!ok) {
            return false;
        }
        index++;
        if (args->size() < index) {
            return false;
        }
        const QString name = args->at(index);
        index++;

        playerList->insert(id, name);
    }
    if (args->size() < index) {
        return false;
    }

    return readLayout(args, index, layout);

}


bool Protocol::parseGameWinner(const QStringList *args, int *gameID, int *playerID)
{

    if (args->size() != 2) {
        return false;
    }

    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }
    *playerID = args->at(1).toInt(&ok);
    return ok;

}



bool Protocol::parseGameChat(const QStringList *args, QString *text, QString *sender, int *gameID)
{

    if (args->size() != 3) {
        return false;
    }

    *text = args->at(0);
    *sender = args->at(1);

    bool ok = false;
    *gameID = args->at(2).toInt(&ok);
    return ok;

}


bool Protocol::parseRequestNick(const QStringList *args, QString *nick)
{

    if (args->size() != 1) {
        return false;
    }
    *nick = args->at(0);
    return true;

}


bool Protocol::parseRequestTurn(const QStringList *args, int *from, int *to, int *gameID)
{

    if (args->size() != 3) {
        return false;
    }

    bool ok = false;
    *from = args->at(0).toInt(&ok);
    if (!ok) {
        return false;
    }
    *to = args->at(1).toInt(&ok);
    if (!ok) {
        return false;
    }
    *gameID = args->at(2).toInt(&ok);
    return ok;

}


bool Protocol::parseCreateGame(const QStringList *args, QString *name, QString *key, QF::Layout *layout)
{

    if (args->size() < 2) {
        return false;
    }

    *name = args->at(0);
    *key = args->at(1);

    return readLayout(args, 2, layout);

}


bool Protocol::parseJoinGame(const QStringList *args, int *gameID, QString *key)
{

    if (args->size() != 2) {
        return false;
    }

    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    *key = args->at(1);

    return ok;

}


bool Protocol::parseStartGameRequest(const QStringList *args, int *gameID)
{

    if (args->size() != 1) {
        return false;
    }

    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    return ok;

}


bool Protocol::parseRequestRoll(const QStringList *args, int *gameID)
{

    if (args->size() != 1) {
        return false;
    }

    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    return ok;

}


bool Protocol::parseLeave(const QStringList *args, int *gameID)
{

    if (args->size() != 1) {
        return false;
    }

    bool ok = false;
    *gameID = args->at(0).toInt(&ok);
    return ok;

}



} // namespace QF
