/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


#ifndef PROTOCOL_H
#define PROTOCOL_H

// own
#include "QFrustrationLib_global.h"
#include "layout.h"

// Qt
#include <QtCore/QStringList>


namespace QF {


class Stone;
class QFRUSTRATIONLIBSHARED_EXPORT Protocol
{


public:
    enum BlobType {
        InvalidType = -1,
        PingType = 0,
        PongType = 1,
        SetNickType = 4,
        ErrorReplyType = 5,
        GameStartType = 9,
        StoneAddedType = 11,
        RollType = 12,
        TurnType = 13,
        StoneMovedType = 14,
        GameJoinedType = 20,
        GameLeftType = 21,
        ErrorType = 22,
        GamePlayerChangedType = 23,
        GameInviteType = 24,
        CreateGameType = 25,
        JoinGameTye = 26,
        StartGameRequestType = 27,
        YouJoinedGameType = 28,
        GameListType = 29,
        GameOpenType = 30,
        GameClosedType = 31,
        GameChatType = 32,
        GameWinnerType = 33
    };

    enum Error {
        TooManyPlayersErrorType = 0,
        NotGameOwnerErrorType = 1,
        NotEnoughPlayersErrorType = 2,
        NotYourTurnErrorType = 3,
        InvalidTurnErrorType = 4,
        GameAlreadyStartedErrorType = 5,
        AlreadyInGameErrorType = 6,
        GameNotFoundErrorType = 7,
        WrongKeyErrorType = 8,
        GameNotStartedErrorType = 9,
        NoNickSetErrorType = 10,
        NickAlreadyInUseErrorType = 11,
        PlayerHasAlreadyRolledErrorType = 12,
        PlayerHasNotRolledErrorType = 13,
        InvalidDataErrorType = 14
    };


    // server + client
    static bool getBlobFromBuffer(QByteArray *buffer, QByteArray *blob);
    static QByteArray createBlob(const QF::Protocol::BlobType &type, const QStringList &args);
    static bool parseBlob(QByteArray &blob, QF::Protocol::BlobType *type, QStringList *args);

    // server
    static QByteArray ping();
    static QByteArray setNick(const QString &nick);
    static QByteArray sendError(const QF::Protocol::BlobType &error);
    static QByteArray sendGameStarted(const QStringList &playerList, const QString &gameID);
    static QByteArray sendStoneAdded(QF::Stone *stone);
    static QByteArray sendRoll(const int &value, const int &gameID);
    static QByteArray sendTurn(const QHash<int, int> &turn, const int &gameID);
    static QByteArray sendStoneMoved(const int &from, const int &to, const int &gameID);
    static QByteArray sendGameLeft(const QString &player, const int &id, const int &gameID);
    static QByteArray sendError(const QF::Protocol::Error &error);
    static QByteArray sendPlayerChanged(const int &playerID, const int &gameID);
    static QByteArray sendGameStarted(const int &gameID);
    static QByteArray sendGameInvite(const QString &from, const int &gameID);
    static QByteArray sendGameCreated(const int &gameID, const QString &name, const QString &key, const QF::Layout &layout);
    static QByteArray sendGameJoined(const QString &playerName, const int &playerID, const int &gameID, const QString &gameName, const QHash<int, QString> &playerList, const QF::Layout &layout);
    static QByteArray sendGameList(const QHash<int, QString> &list);
    static QByteArray sendGameOpen(const int &gameID, const QString &name);
    static QByteArray sendGameClosed(const int &gameID, const QString &name);
    static QByteArray sendGameWinner(const int &gameID, const int &playerID);

    // client
    static QByteArray pong();
    static QByteArray requestNick(const QString &nick);
    static QByteArray requestRoll(const int &gameID);
    static QByteArray requestTurn(const int &from, const int &to, const int &gameID);
    static QByteArray sendCreateGame(const QString &name, const QString &key, const QF::Layout &layout);
    static QByteArray requestGameList();
    static QByteArray requestJoinGame(const int &gameID, const QString &key);
    static QByteArray requestStart(const int &gameID);
    static QByteArray requestLeave(const int &gameID);

    // client parse
    static bool parseServerSetsNick(const QStringList *args, QString *nick);
    static bool parseError(const QStringList *args, QF::Protocol::Error *error);
    static bool parseGameStarted(const QStringList *args, int *gameID);
    static bool parseCurrentPlayerChanged(const QStringList *args, int *playerID, int *gameID);
    static bool parseStoneAdded(const QStringList *args, int *playerID, int *stoneID, int *stonePos, int *gameID);
    static bool parseRoll(const QStringList *args, int *value, int *gameID);
    static bool parseTurn(const QStringList *args, QHash<int, int> *turn, int *gameID);
    static bool parseStoneMoved(const QStringList *args, int *from, int *to, int *gameID);
    static bool parseGameCreated(const QStringList *args, int *gameID, QString *name, QString *key, QF::Layout *layout);
    static bool parseGameLeft(const QStringList *args, QString *nick, int *playerID, int *gameID);
    static bool parseGameJoined(const QStringList *args, QString *playerName, int *playerID, int *gameID, QString *gameName, QHash<int, QString> *playerList, QF::Layout *layout);
    static bool parseGameList(const QStringList *args, QHash<int, QString> *gameList);
    static bool parseGameOpen(const QStringList *args, int *gameID, QString *name);
    static bool parseGameClosed(const QStringList *args, int *gameID, QString *name);
    static bool parseGameWinner(const QStringList *args, int *gameID, int *playerID);

    // client + server
    static QByteArray sendGameChat(const QString &text, const QString &sender, const int &gameID);
    static bool parseGameChat(const QStringList *args,QString *text, QString *sender, int *gameID);

    // server parse
    static bool parseRequestNick(const QStringList *args, QString *nick);
    static bool parseRequestTurn(const QStringList *args, int *from, int *to, int *gameID);
    static bool parseCreateGame(const QStringList *args, QString *name, QString *key, QF::Layout *layout);
    static bool parseJoinGame(const QStringList *args, int *gameID, QString *key);
    static bool parseStartGameRequest(const QStringList *args, int *gameID);
    static bool parseRequestRoll(const QStringList *args, int *gameID);
    static bool parseLeave(const QStringList *args, int *gameID);

};


} // namespace QF


#endif // PROTOCOL_H
