/***************************************************************************
 *   Copyright (C) 2010 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "server.h"
#include "server_p.h"
#include "client.h"
#include "game.h"

// lib
#include <QFrustrationLib/protocol.h>

// Qt
#include <QtCore/QDebug>
#include <QtCore/QStringList>


namespace QFS {


ServerPrivate::ServerPrivate(Server *parent)
    : QTcpServer(parent),
    q(parent)
{


}


ServerPrivate::~ServerPrivate()
{

    qDeleteAll(clients);

}


void ServerPrivate::sendToAllClients(const QByteArray &blob)
{

    for (int i = 0; i < clients.size(); i++) {
        Client *client = clients.at(i);

        client->send(blob);
    }

}


void ServerPrivate::clientConnected(QFS::Client *self)
{

    emit q->clientConnected(self->nick());

}


void ServerPrivate::clientDisconnected(QFS::Client *client)
{

    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *game = m_games.at(i);

        if (game->clients().contains(client)) {
            game->removePlayer(client);
        }
    }

    if (!client->nick().isEmpty()) {
        emit q->clientDisconnected(client->nick());
    }

    clients.removeAll(client);
    client->deleteLater();

}


void ServerPrivate::clientRollRequested(QFS::Client *self, const int &gameID)
{

    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *game = m_games.at(i);

        if (game->gameId() == gameID) {
            game->createRoll(self);
            break;
        }
    }

}


void ServerPrivate::clientTurnRequested(QFS::Client *self, const int &from, const int &to, const int &gameID)
{

    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *game = m_games.at(i);

        if (game->gameId() == gameID) {
            game->setTurn(self, from, to);
            break;
        }
    }

}


void ServerPrivate::gameCreateRequested(QFS::Client *self, const QString &name, const QString &key, const QF::Layout &layout)
{

    int id = 1;
    while (true) {
        bool found = false;
        for (int i = 0; i < m_games.size(); i++) {
            QFS::Game *game = m_games.at(i);
            if (game->gameId() == id) {
                found = true;
                break;
            }
        }

        if (found) {
            found = false;
            id++;
        } else {
            break;
        }
    }

    QFS::Game *game = new QFS::Game(self, id, name, key, layout, this);
    connect(game, SIGNAL(gameClosed(QFS::Game*)), this, SLOT(gameClosed(QFS::Game*)));
    connect(game, SIGNAL(gameStarted(QFS::Game*)), this, SLOT(gameStarted(QFS::Game*)));
    m_games.append(game);

    sendToAllClients(QF::Protocol::sendGameOpen(id, name));

    self->send(QF::Protocol::sendGameCreated(id, name, key, layout));

    emit q->gameCreated(name);

}


void ServerPrivate::joinGameRequested(QFS::Client *self, const int &gameID, const QString &key)
{

    QFS::Game *game = 0;
    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *g = m_games.at(i);

        if (g->clients().contains(self)) {
            self->send(QF::Protocol::sendError(QF::Protocol::AlreadyInGameErrorType));
            return;
        }

        if (g->gameId() == gameID) {
            game = g;
        }
    }

    if (!game) {
        self->send(QF::Protocol::sendError(QF::Protocol::GameNotFoundErrorType));
        return;
    }

    if (game->key() != key) {
        self->send(QF::Protocol::sendError(QF::Protocol::WrongKeyErrorType));
        return;
    }

    game->addPlayer(self);

}


void ServerPrivate::startGameRequested(QFS::Client *self, const int &gameID)
{

    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *game = m_games.at(i);

        if (game->gameId() == gameID) {
            game->startGame(self);
            return;
        }
    }
    self->send(QF::Protocol::sendError(QF::Protocol::GameNotFoundErrorType));

}


void ServerPrivate::gameListRequested(QFS::Client *self)
{

    QHash<int, QString> list;
    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *game = m_games.at(i);

        list.insert(game->gameId(), game->name());
    }
    self->send(QF::Protocol::sendGameList(list));

}


void ServerPrivate::gameClosed(QFS::Game *self)
{

    emit q->gameClosed(self->name());

    m_games.removeAll(self);
    sendToAllClients(QF::Protocol::sendGameClosed(self->gameId(), self->name()));
    self->deleteLater();

}


void ServerPrivate::newGameChatText(QFS::Client *self, const QString &text, const QString &sender, const int &gameID)
{

    Q_UNUSED(self);
    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *game = m_games.at(i);

        if (game->gameId() == gameID) {
            game->sendChat(sender, text);
            break;
        }
    }

}


void ServerPrivate::leaveRequested(QFS::Client *self, const int &gameID)
{

    for (int i = 0; i < m_games.size(); i++) {
        QFS::Game *game = m_games.at(i);

        if (game->gameId() == gameID) {
            game->removePlayer(self);
            break;
        }
    }

}


void ServerPrivate::gameStarted(QFS::Game *self)
{

    emit q->gameStarted(self->name());

}


void ServerPrivate::incomingConnection(int socketDescriptor)
{

    Client *client = new Client(this, q);
    if (client->init(socketDescriptor)) {
        connect(client, SIGNAL(disconnected(QFS::Client*)), this, SLOT(clientDisconnected(QFS::Client*)));
        connect(client, SIGNAL(gameCreateRequested(QFS::Client*,QString,QString,QF::Layout)), this,
                SLOT(gameCreateRequested(QFS::Client*,QString,QString,QF::Layout)));
        connect(client, SIGNAL(joinGameRequested(QFS::Client*,int,QString)), this, SLOT(joinGameRequested(QFS::Client*,int,QString)));
        connect(client, SIGNAL(startGameRequested(QFS::Client*,int)), this, SLOT(startGameRequested(QFS::Client*,int)));
        connect(client, SIGNAL(gameListRequested(QFS::Client*)), this, SLOT(gameListRequested(QFS::Client*)));
        connect(client, SIGNAL(turnRequested(QFS::Client*,int,int,int)), this, SLOT(clientTurnRequested(QFS::Client*,int,int,int)));
        connect(client, SIGNAL(rollRequested(QFS::Client*,int)), this, SLOT(clientRollRequested(QFS::Client*,int)));
        connect(client, SIGNAL(newGameChatText(QFS::Client*,QString,QString,int)), this, SLOT(newGameChatText(QFS::Client*,QString,QString,int)));
        connect(client, SIGNAL(leaveRequested(QFS::Client*,int)), this, SLOT(leaveRequested(QFS::Client*,int)));
        connect(client, SIGNAL(connected(QFS::Client*)), this, SLOT(clientConnected(QFS::Client*)));
        connect(client, SIGNAL(receivingData(qint64)), q, SIGNAL(receivingData(qint64)));
        connect(client, SIGNAL(dataSent(qint64)), q, SIGNAL(dataSent(qint64)));

        clients.append(client);
    } else {
        delete client;
    }

}



Server::Server(QObject *parent)
    : QObject(parent),
    d(new ServerPrivate(this))
{



}


Server::~Server()
{

    delete d;

}


quint16 Server::port() const
{

    if (d->isListening()) {
        return d->serverPort();
    }
    return d->port;

}


QString Server::hostName() const
{

    return d->serverAddress().toString();

}


QStringList Server::players() const
{

    QStringList list;
    for (int i = 0; i < d->clients.size(); i++) {
        const QString nick = d->clients.at(i)->nick();
        if (!nick.isEmpty()) {
            list.append(nick);
        }
    }
    return list;

}


bool Server::isNickValid(const QString &nick)
{

    return !players().contains(nick);

}


bool Server::start(const bool &publicServer)
{

    return d->listen(publicServer ? QHostAddress::Any : QHostAddress::LocalHost, d->port);

}


QString Server::address() const
{

    return d->serverAddress().toString();

}


QString Server::errorString() const
{

    return d->errorString();

}


void Server::setPort(const quint16 &port)
{

    d->port = port;

}


} // namespace QFS
